'''

A scenario discovery oriented implementation of PRIM.

The implementation of prim provided here is datatype aware, so
categorical variables will be handled appropriately. It also uses a 
non-standard objective function in the peeling and pasting phase of the
algorithm. This algorithm looks at the increase in the mean divided 
by the amount of data removed. So essentially, it uses something akin
to the first order derivative of the original objective function. 

The implementation is designed for interactive use in combination with the
ipython notebook. 

'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

import copy
import math
from operator import itemgetter

import matplotlib as mpl
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import numpy.lib.recfunctions as rf
import pandas as pd
import six
from mpl_toolkits.axes_grid1 import host_subplot  # @UnresolvedImport
from scipy.stats import binom
from ema_workbench.analysis import scenario_discovery_util

try:
    import mpld3
except ImportError:
    global mpld3
    mpdl3 = None

from .plotting_util import make_legend, COLOR_LIST
from ..util import info, debug, EMAError

# from . import pairs_plotting
from . import scenario_discovery_util as sdutil

# Created on 22 feb. 2013
# 
# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

# TODO:: make separate qp-test for the lower limit and the upper limit

__all__ = ['ABOVE', 'BELOW', 'setup_prim', 'Prim', 'PrimBox', 'PrimException',
           'MultiBoxesPrim']

LENIENT2 = 'lenient2'
LENIENT1 = 'lenient1'
ORIGINAL = 'original'

ABOVE = 1
BELOW = -1
PRECISION = '.2f'

def get_quantile(data, quantile):
    '''
    quantile calculation modeled on the implementation used in sdtoolkit

    Parameters
    ----------
    data : nd array like 
           dataset for which quantile is needed
    quantile : float
               the desired quantile
    
    '''
    assert quantile>0
    assert quantile<1
 
    data = data.compressed()
    data = list(data) #TODO do we really need a list? I doubt it
    data.sort()    
    
    i = (len(data)-1)*quantile
    index_lower =  int(math.floor(i))
    index_higher = int(math.ceil(i))
    
    value = 0

    if quantile > 0.5:
        # upper
        while (data[index_lower] == data[index_higher]) & (index_lower>0):
            index_lower -= 1
        value = (data[index_lower]+data[index_higher])/2
    else:
        #lower
        while (data[index_lower] == data[index_higher]) & \
              (index_higher<len(data)-1):
            index_higher += 1
        value = (data[index_lower]+data[index_higher])/2

    return value


def _pair_wise_scatter(x,y, box_lim, restricted_dims):
    ''' helper function for pair wise scatter plotting
    
    #TODO the cases of interest should be in red rather than in blue
    # this will give a nice visual insight into the quality of the box
    # currently it is done through the face color being white or blue
    # this is not very clear
    
    
    Parameters
    ----------
    x : numpy structured array
        the experiments
    y : numpy array
        the outcome of interest
    box_lim : numpy structured array
              a boxlim
    restricted_dims : list of strings
                      list of uncertainties that define the boxlims
    
    '''

    restricted_dims = list(restricted_dims)
    combis = [(field1, field2) for field1 in restricted_dims\
                               for field2 in restricted_dims]

    grid = gridspec.GridSpec(len(restricted_dims), len(restricted_dims))                             
    grid.update(wspace = 0.1,
                hspace = 0.1)    
    figure = plt.figure()
    
    for field1, field2 in combis:
        i = restricted_dims.index(field1)
        j = restricted_dims.index(field2)
        ax = figure.add_subplot(grid[i,j])        
        ec='b'
        fc='b'
        
        if field1==field2:
            ec='white'
            fc='white'
        
        # scatter points
        for n in [0,1]:
            x_n = x[y==n]        
            x_1 = x_n[field2]
            x_2 = x_n[field1]
            
            if (n==0) :
                fc='white'
            elif ec=='b':
                fc='b'
            
            ax.scatter(x_1, x_2, facecolor=fc, edgecolor=ec, s=10)

        # draw boxlim
        if field1 != field2:
            x_1 = box_lim[field2]
            x_2 = box_lim[field1]

            for n in [0,1]:
                ax.plot(x_1,
                    [x_2[n], x_2[n]], c='r', linewidth=3)
                ax.plot([x_1[n], x_1[n]],
                    x_2, c='r', linewidth=3)
            
#         #reuse labeling function from pairs_plotting
#         pairs_plotting.do_text_ticks_labels(ax, i, j, field1, field2, None, 
#                                             restricted_dims)
            
    return figure

class CurEntry(object):
    '''a descriptor for the current entry on the peeling and pasting 
    trajectory'''
    
    def __init__(self, name):
        self.name = name
        
    def __get__(self, instance, owner):
        return instance.peeling_trajectory[self.name][instance._cur_box]
    
    def __set__(self, instance, value):
        raise PrimException("this property cannot be assigned to")
                

class PrimBox(object):
    '''A class that holds information over a specific box 
    
    Attributes
    ----------
    coverage : float
               coverage of currently selected box
    density : float
               density of currently selected box
    mean : float
           mean of currently selected box
    res_dim : int
              number of restricted dimensions of currently selected box
    mass : float
           mass of currently selected box 
    peeling_trajectory : pandas dataframe
                         stats for each box in peeling trajectory
    box_lims : list
               list of box lims for each box in peeling trajectory

    
    by default, the currently selected box is the last box on the peeling
    trajectory, unless this is changed via :meth:`PrimBox.select`.
    
    '''
    
    coverage = CurEntry('coverage')
    density = CurEntry('density')
    mean = CurEntry('mean')
    res_dim = CurEntry('res dim')
    mass = CurEntry('mass')
    
    _frozen=False
    
    def __init__(self, prim, box_lims, indices):
        '''init 
        
        Parameters
        ----------
        prim : Prim instance
        box_lims : recarray
        indices : ndarray
        
        
        '''
        
        self.prim = prim
        
        # peeling and pasting trajectory
        colums = ['coverage', 'density', 'mean', 'res dim', 'mass']
        self.peeling_trajectory = pd.DataFrame(columns=colums)
        
        self.box_lims = []
        self._cur_box = -1
        
        # indices van data in box
        self.update(box_lims, indices)

    def __getattr__(self, name):
        '''
        used here to give box_lim same behaviour as coverage, density, mean
        res_dim, and mass. That is, it will return the box lim associated with
        the currently selected box. 
        '''
        
        if name=='box_lim':
            return self.box_lims[self._cur_box]
        else:
            raise AttributeError

    def inspect(self, i=None, style='table'):
        '''
        
        Write the stats and box limits of the user specified box to standard 
        out. if i is not provided, the last box will be printed
        
        Parameters
        ----------
        i : int, optional
            the index of the box, defaults to currently selected box
        style : {'table', 'graph'}
                the style of the visualization
        
        '''
        if i == None:
            i = self._cur_box
        
        stats = self.peeling_trajectory.iloc[i].to_dict()
        stats['restricted_dim'] = stats['res dim']

        qp_values = self._calculate_quasi_p(i)
        uncs = [(key, value) for key, value in qp_values.items()]
        uncs.sort(key=itemgetter(1))
        uncs = [uncs[0] for uncs in uncs]
        
        if style == 'table':
            return self._inspect_table(i, uncs, qp_values)
        elif style == 'graph':
            return self._inspect_graph(i, uncs, qp_values)
        else:
            raise ValueError("style must be one of graph or table")
            
    def _inspect_table(self, i, uncs, qp_values):
        '''Helper function for visualizing box statistics in 
        table form'''
        #make the descriptive statistics for the box
        print(self.peeling_trajectory.iloc[i])
        print()
        
        # make the box definition
        columns = pd.MultiIndex.from_product([['box {}'.format(i)],
                                              ['min', 'max', 'qp values']])
        box_lim = pd.DataFrame(np.zeros((len(uncs), 3)), 
                               index=uncs, 
                               columns=columns)
        
        for unc in uncs:
            values = self.box_lims[i][unc][:]
            box_lim.loc[unc] = [values[0], values[1], qp_values[unc]]
        
        print(box_lim)
        print()
        
    def _inspect_graph(self,  i, uncs, qp_values):
        '''Helper function for visualizing box statistics in 
        graph form'''        
        
        # normalize the box lims
        # we don't need to show the last box, for this is the 
        # box_init, which is visualized by a grey area in this
        # plot.
        box_lim_init = self.prim.box_init
        box_lim = self.box_lims[i]
        norm_box_lim =  sdutil._normalize(box_lim, box_lim_init, uncs)
        
        fig, ax = sdutil._setup_figure(uncs)
        color = COLOR_LIST[0]
        for j, u in enumerate(uncs):
            # we want to have the most restricted dimension
            # at the top of the figure
            xj = len(uncs) - j - 1

            self.prim._plot_unc(box_lim_init, xj, j, 0, norm_box_lim, box_lim, 
                                u, ax, color)

            # new part
            dtype = box_lim_init[u].dtype
            
            props = {'facecolor':'white',
                     'edgecolor':'white',
                     'alpha':0.25}
            y = xj

        
            if dtype == object:
                pass
                elements = sorted(list(box_lim_init[u][0]))
                max_value = (len(elements)-1)
                values = box_lim[u][0]
                x = [elements.index(entry) for entry in 
                     values]
                x = [entry/max_value for entry in x]
                
                for xi, label in zip(x, values):
                    ax.text(xi, y-0.1, label, ha='center', va='center',
                           bbox=props, color='blue', fontweight='normal')

            else:
                props = {'facecolor':'white',
                         'edgecolor':'white',
                         'alpha':0.25}
    
                # plot limit text labels
                x = norm_box_lim[j][0]
    
                if not np.allclose(x, 0):
                    label = "{: .2g}".format(self.box_lims[i][u][0])
                    ax.text(x, y-0.2, label, ha='center', va='center',
                           bbox=props, color='blue', fontweight='normal')
    
                x = norm_box_lim[j][1]
                if not np.allclose(x, 1):
                    label = "{: .2g}".format(self.box_lims[i][u][1])
                    ax.text(x, y-0.2, label, ha='center', va='center',
                           bbox=props, color='blue', fontweight='normal')

                # plot uncertainty space text labels
                x = 0
                label = "{: .2g}".format(box_lim_init[u][0])
                ax.text(x-0.01, y, label, ha='right', va='center',
                       bbox=props, color='black', fontweight='normal')
    
                x = 1
                label = "{: .2g}".format(box_lim_init[u][1])
                ax.text(x+0.01, y, label, ha='left', va='center',
                       bbox=props, color='black', fontweight='normal')
                
            # set y labels
            labels = ["{} ({:.2g})".format(u, qp_values[u]) for u in uncs]
            labels = labels[::-1]
            ax.set_yticklabels(labels)

            # remove x tick labels
            ax.set_xticklabels([])

            # add table to the left
            coverage = '{:.3g}'.format(self.peeling_trajectory['coverage'][i])
            density = '{:.3g}'.format(self.peeling_trajectory['density'][i])
            
            ax.table(cellText=[[coverage], [density]],
                    colWidths = [0.1]*2,
                    rowLabels=['coverage', 'density'],
                    colLabels=None,
                    loc='right',
                    bbox=[1.2, 0.9, 0.1, 0.1])
        
#             plt.tight_layout()
        return fig
        
    def select(self, i):
        '''        
        select an entry from the peeling and pasting trajectory and update
        the prim box to this selected box.
        
        Parameters
        ----------
        i : int
            the index of the box to select.
        
        '''
        if self._frozen:
            raise PrimException("""box has been frozen because PRIM has found 
                                at least one more recent box""")
        
        indices = sdutil._in_box(self.prim.x[self.prim.yi_remaining], 
                                 self.box_lims[i])
        self.yi = self.prim.yi_remaining[indices]
        self._cur_box = i

    def drop_restriction(self, uncertainty):
        '''
        drop the restriction on the specified dimension. That is, replace
        the limits in the chosen box with a new box where for the specified 
        uncertainty the limits of the initial box are being used. The resulting
        box is added to the peeling trajectory.
        
        Parameters
        ----------
        uncertainty : str
        
        '''
        
        new_box_lim = copy.deepcopy(self.box_lim)
        new_box_lim[uncertainty][:] = self.box_lims[0][uncertainty][:]
        indices = sdutil._in_box(self.prim.x[self.prim.yi_remaining], 
                                 new_box_lim)
        indices = self.prim.yi_remaining[indices]
        self.update(new_box_lim, indices)
        
    def update(self, box_lims, indices):
        '''
        
        update the box to the provided box limits.
        
        Parameters
        ----------
        box_lims: numpy recarray
                  the new box_lims
        indices: ndarray
                 the indices of y that are inside the box
      
        '''
        self.yi = indices
        
        y = self.prim.y[self.yi]

        self.box_lims.append(box_lims)

        coi = self.prim.determine_coi(self.yi)

        data = {'coverage':coi/self.prim.t_coi, 
                'density':coi/y.shape[0],  
                'mean':np.mean(y),
                'res dim':sdutil._determine_nr_restricted_dims(self.box_lims[-1], 
                                                              self.prim.box_init),
                'mass':y.shape[0]/self.prim.n}
        new_row = pd.DataFrame([data])
        self.peeling_trajectory = self.peeling_trajectory.append(new_row, 
                                                             ignore_index=True)
        
        self._cur_box = len(self.peeling_trajectory)-1
        
    def show_ppt(self):
        '''show the peeling and pasting trajectory in a figure'''
        
        ax = host_subplot(111)
        ax.set_xlabel("peeling and pasting trajectory")
        
        par = ax.twinx()
        par.set_ylabel("nr. restricted dimensions")
            
        ax.plot(self.peeling_trajectory['mean'], label="mean")
        ax.plot(self.peeling_trajectory['mass'], label="mass")
        ax.plot(self.peeling_trajectory['coverage'], label="coverage")
        ax.plot(self.peeling_trajectory['density'], label="density")
        par.plot(self.peeling_trajectory['res dim'], label="restricted dims")
        ax.grid(True, which='both')
        ax.set_ylim(ymin=0,ymax=1)
        
        fig = plt.gcf()
        
        make_legend(['mean', 'mass', 'coverage', 'density', 'restricted_dim'],
                    ax, ncol=5, alpha=1)
        return fig
    
    def show_tradeoff(self):
        '''Visualize the trade off between coverage and density. Color is used
        to denote the number of restricted dimensions.'''
       
        fig = plt.figure()
        ax = fig.add_subplot(111, aspect='equal')
        
        cmap = mpl.cm.YlGnBu_r #@UndefinedVariable
        boundaries = np.arange(-0.5, 
                               max(self.peeling_trajectory['res dim'])+1.5, 
                               step=1)
        ncolors = cmap.N
        norm = mpl.colors.BoundaryNorm(boundaries, ncolors)
        
        p = ax.scatter(self.peeling_trajectory['coverage'], 
                       self.peeling_trajectory['density'], 
                       c=self.peeling_trajectory['res dim'], 
                       norm=norm,
                       cmap=cmap)
        ax.set_ylabel('density')
        ax.set_xlabel('coverage')
        ax.set_ylim(ymin=0, ymax=1.2)
        ax.set_xlim(xmin=0, xmax=1.2)
        
        ticklocs = np.arange(0, 
                             max(self.peeling_trajectory['res dim'])+1, 
                             step=1)
        cb = fig.colorbar(p, spacing='uniform', ticks=ticklocs, drawedges=True)
        cb.set_label("nr. of restricted dimensions")
        
        # make the tooltip tables
        if mpld3:
            # Define some CSS to control our custom labels
            css = """
            table
            {
              border-collapse: collapse;
            }
            th
            {
              background-color:  rgba(255,255,255,0.6);;;
            }
            td
            {
              background-color: rgba(255,255,255,0.6);;
            }
            table, th, td
            {
              font-family:Tahoma, Tahoma, sans-serif;
              font-size: 16px;
              border: 1px solid black;
              text-align: right;
            }
            """   
            
            labels = []
            columns_to_include = ['coverage','density', 'mass', 'res dim']
            frmt = lambda x: '{:.2f}'.format( x )
            for i in range(len(self.peeling_trajectory['coverage'])):
                label = self.peeling_trajectory.ix[[i], columns_to_include].T
                label.columns = ['box {0}'.format(i)]
                # .to_html() is unicode; so make leading 'u' go away with str()
                labels.append(str(label.to_html(float_format=frmt)))        
    
            tooltip = mpld3.plugins.PointHTMLTooltip(p, labels, voffset=10, 
                                               hoffset=10, css=css)  
            mpld3.plugins.connect(fig, tooltip)        
        
        return fig
    
    def show_pairs_scatter(self):
        '''
        
        make a pair wise scatter plot of all the restricted dimensions
        with color denoting whether a given point is of interest or not
        and the boxlims superimposed on top.
        
        '''   
        return _pair_wise_scatter(self.prim.x, self.prim.y, self.box_lim, 
                           sdutil._determine_restricted_dims(self.box_lim, 
                                                        self.prim.box_init))
    
    def write_ppt_to_stdout(self):
        '''write the peeling and pasting trajectory to stdout'''
        print(self.peeling_trajectory)
        print("\n")

    def _calculate_quasi_p(self, i):
        '''helper function for calculating quasi-p values as discussed in 
        Bryant and Lempert (2010). This is a one sided  binomial test. 
        
        Parameters
        ----------
        i : int
            the specific box in the peeling trajectory for which the quasi-p 
            values are to be calculated.
        
        Returns
        -------
        dict
        
        '''
        
        box_lim = self.box_lims[i]
        restricted_dims = list(sdutil._determine_restricted_dims(box_lim,
                                                           self.prim.box_init))
        
        # total nr. of cases in box
        Tbox = self.peeling_trajectory['mass'][i] * self.prim.n 
        
        # total nr. of cases of interest in box
        Hbox = self.peeling_trajectory['coverage'][i] * self.prim.t_coi  
        
        qp_values = {}
        
        for u in restricted_dims:
            temp_box = copy.deepcopy(box_lim)
            temp_box[u] = self.box_lims[0][u]
        
            indices = sdutil._in_box(self.prim.x[self.prim.yi_remaining], 
                                     temp_box)
            indices = self.prim.yi_remaining[indices]
            
            # total nr. of cases in box with one restriction removed
            Tj = indices.shape[0]  
            
            # total nr. of cases of interest in box with one restriction 
            # removed
            Hj = np.sum(self.prim.y[indices])
            
            p = Hj/Tj
            
            Hbox = int(Hbox)
            Tbox = int(Tbox)
            
            qp = binom.sf(Hbox-1, Tbox, p)
            qp_values[u] = qp
            
        return qp_values

    def _format_stats(self, nr, stats):
        '''helper function for formating box stats'''
        row = self.stats_format.format(nr,**stats)
        return row


class PrimException(Exception):
    '''Base exception class for prim related exceptions'''
    pass


def setup_prim(results, classify, threshold, incl_unc=[], **kwargs):
    """Helper function for setting up the prim algorithm
    
    Parameters
    ----------
    results : tuple
              tuple of structured array and dict with numpy arrays
              the return from :meth:`perform_experiments`.
    classify : str or callable
               either a string denoting the outcome of interest to 
               use or a function. 
    threshold : double
                the minimum score on the objective function of the last box
                on the peeling trajectory. In case of a binary classification,
                this should be between 0 and 1. 
    incl_unc : list of str, optional
               list of uncertainties to include in prim analysis
    kwargs : dict
             valid keyword arguments for prim.Prim
    
    Returns
    -------
    a Prim instance
    
    Raises
    ------
    PrimException 
        if data resulting from classify is not a 1-d array. 
    TypeError 
        if classify is not a string or a callable.
    
    """
    
    if not incl_unc:
        x = np.ma.array(results[0])
    else:
        drop_names = set(rf.get_names(results[0].dtype))-set(incl_unc)
        x = rf.drop_fields(results[0], drop_names, asrecarray=True)
    if isinstance(classify, six.string_types):
        y = results[1][classify]
    elif callable(classify):
        y = classify(results[1])
    else:
        raise TypeError("unknown type for classify")
    
    return Prim(x,y, threshold=threshold, **kwargs)
    

class Prim(sdutil.OutputFormatterMixin):
    '''Patient rule induction algorithm
    
    The implementation of Prim is tailored to interactive use in the context
    of scenario discovery

    Parameters
    ----------
    x : structured array
        the independent variables
    y : 1d ndarray
        the dependent variable
    threshold : float
                the coverage threshold that a box has to meet
    obj_function : {LENIENT1, LENIENT2, ORIGINAL}
                   the objective function used by PRIM. Defaults to a lenient 
                   objective function based on the gain of mean divided by the 
                   loss of mass. 
    peel_alpha : float, optional 
                 parameter controlling the peeling stage (default = 0.05). 
    paste_alpha : float, optional
                  parameter controlling the pasting stage (default = 0.05).
    mass_min : float, optional
               minimum mass of a box (default = 0.05). 
    threshold_type : {ABOVE, BELOW}
                     whether to look above or below the threshold value
  
        
    See also
    --------
    :mod:`cart`
    
    
    '''
    
    message = "{0} points remaining, containing {1} cases of interest"
    
    def __init__(self, 
                 x,
                 y, 
                 threshold, 
                 obj_function=LENIENT1, 
                 peel_alpha=0.05, 
                 paste_alpha=0.05,
                 mass_min=0.05, 
                 threshold_type=ABOVE):
        
        # preprocess x
        x = rf.drop_fields(x, "scenario_id", asrecarray=True)
        x = np.ma.array(x)
        names = rf.get_names(x.dtype)
        for name in names:
            logical = pd.isnull(x[name])
            x[name][logical] = np.ma.masked
        
        self.x = x
        self.y = y
        
        if len(self.y.shape) > 1:
            raise PrimException("y is not a 1-d array")
        
        # store the remainder of the parameters
        self.paste_alpha = paste_alpha
        self.peel_alpha = peel_alpha
        self.mass_min = mass_min
        self.threshold = threshold 
        self.threshold_type = threshold_type
        self.obj_func = self._obj_functions[obj_function]
       
        # set the indices
        self.yi = np.arange(0, self.y.shape[0])
       
        # how many data points do we have
        self.n = self.y.shape[0]
        
        # how many cases of interest do we have?
        self.t_coi = self.determine_coi(self.yi)
        
        # initial box that contains all data
        self.box_init = sdutil._make_box(self.x)
    
        # make a list in which the identified boxes can be put
        self._boxes = []
        
        self._update_yi_remaining()
    
    @property
    def boxes(self):
        boxes = [box.box_lim for box in self._boxes]
        
        if not boxes:
            return [self.box_init]
#         elif not np.all(sdutil._compare(boxes[-1], self.box_init)):
#                 boxes.append(self.box_init)
        return boxes 
    
    @property
    def stats(self):
        stats = []
        items = ['coverage','density', 'mass', 'res_dim']
        for box in self._boxes:
            stats.append({key: getattr(box, key) for key in items})
        return stats
    
    def perform_pca(self, subsets=None, exclude=set()):
        '''
        
        WARNING:: code still needs to be tested!!!
        
        Pre-process the data by performing a pca based rotation on it. 
        This effectively turns the algorithm into PCA-PRIM as described
        in `Dalal et al (2013) <http://www.sciencedirect.com/science/article/pii/S1364815213001345>`_
        
        Parameters
        ----------
        subsets: dict, optional 
                 expects a dictionary with group name as key and a list of 
                 uncertainty names as values. If this is used, a constrained 
                 PCA-PRIM is executed 
                 
                ..note:: the list of uncertainties should not contain 
                         categorical uncertainties. 
        exclude : list of str, optional 
                  the uncertainties that should be excluded from the rotation
        
        '''
        
        #transform experiments to numpy array
        dtypes = self.x.dtype.fields
        object_dtypes = [key for key, value in dtypes.items() 
                         if value[0]==np.dtype(object)]
        
        #get experiments of interest
        # TODO this assumes binary classification!!!!!!!
        logical = self.y>=self.threshold
        
        # if no subsets are provided all uncertainties with non dtype object 
        # are in the same subset, the name of this is r, for rotation
        if not subsets:
            subsets = {"r":[key for key, value in dtypes.items() 
                            if value[0].name!=np.dtype(object)]}
        else:
            # remove uncertainties that are in exclude and check whether 
            # uncertainties occur in more then one subset
            seen = set()
            for key, value in subsets.items():
                value = set(value) - set(exclude)
    
                subsets[key] = list(value)
                if (seen & value):
                    raise EMAError("uncertainty occurs in more then one subset")
                else:
                    seen = seen | set(value)
            
        #prepare the dtypes for the new rotated experiments recarray
        new_dtypes = []
        for key, value in subsets.items():
            self._assert_dtypes(value, dtypes)
            
            # the names of the rotated columns are based on the group name 
            # and an index
            [new_dtypes.append((str("{}_{}".format(key, i)), float)) for i 
             in range(len(value))]
        
        #add the uncertainties with object dtypes to the end
        included_object_dtypes = set(object_dtypes)-set(exclude)
        [new_dtypes.append((name, object)) for name in included_object_dtypes]
        
        #make a new empty recarray
        rotated_experiments = np.empty((self.x.shape[0],), dtype=new_dtypes)
        
        #put the uncertainties with object dtypes already into the new recarray 
        for name in included_object_dtypes :
            rotated_experiments[name] = self.x[name]
        
        #iterate over the subsets, rotate them, and put them into the new 
        # recarray
        shape = 0
        for key, value in subsets.items():
            shape += len(value) 
        rotation_matrix = np.zeros((shape,shape))
        column_names = []
        row_names = []
        
        j = 0
        for key, value in subsets.items():
            data = self._rotate_subset(value, self.x, logical)
            subset_rotation_matrix, subset_experiments = data 
            rotation_matrix[j:j+len(value), j:j+len(value)] = subset_rotation_matrix
            [row_names.append(entry) for entry in value]
            j += len(value)
            
            for i in range(len(value)):
                name = "%s_%s" % (key, i)
                rotated_experiments[name] = subset_experiments[:,i]
                [column_names.append(name)]
        
        self.rotation_matrix = rotation_matrix
        self.column_names = column_names
        self.row_names = row_names
        
        self.x = np.ma.array(rotated_experiments)
        self.box_init = sdutil._make_box(self.x)
    
    def find_box(self):
        '''Execute one iteration of the PRIM algorithm. That is, find one
        box, starting from the current state of Prim.'''
        # set the indices
        self._update_yi_remaining()
        
        # make boxes already found immutable 
        for box in self._boxes:
            box._frozen = True
        
        if self.yi_remaining.shape[0] == 0:
            info("no data remaining")
            return
        
        # log how much data and how many coi are remaining
        info(self.message.format(self.yi_remaining.shape[0],
                                 self.determine_coi(self.yi_remaining)))
        
        # make a new box that contains all the remaining data points
        box = PrimBox(self, self.box_init, self.yi_remaining[:])
        
        #  perform peeling phase
        box = self._peel(box)
        debug("peeling completed")

        # perform pasting phase        
        box = self._paste(box)
        debug("pasting completed")
        
        message = "mean: {0}, mass: {1}, coverage: {2}, density: {3} restricted_dimensions: {4}"
        message = message.format(box.mean,
                                 box.mass,
                                 box.coverage,
                                 box.density,
                                 box.res_dim)

        if (self.threshold_type==ABOVE) &\
           (box.mean >= self.threshold):
            info(message)
            self._boxes.append(box)
            return box
        elif (self.threshold_type==BELOW) &\
           (box.mean <= self.threshold):
            info(message)
            self._boxes.append(box)
            return box
        else:
            # make a dump box
            info('box does not meet threshold criteria, value is {}, returning dump box'.format(box.mean))
            box = PrimBox(self, self.box_init, self.yi_remaining[:])
            self._boxes.append(box)
            return box

    def determine_coi(self, indices):
        '''        
        Given a set of indices on y, how many cases of interest are there in 
        this set.
        
        Parameters
        ----------
        indices: ndarray
                 a valid index for y

        Returns
        ------- 
        int
            the number of cases of interest.
        
        Raises
        ------
        ValueError 
            if threshold_type is not either ABOVE or BELOW

        '''
        
        y = self.y[indices]
        
        if self.threshold_type == ABOVE:
            coi = y[y >= self.threshold].shape[0]
        elif self.threshold_type == BELOW:
            coi = y[y <= self.threshold].shape[0]
        else:
            raise ValueError("threshold type is not one of ABOVE or BELOW")
        
        return coi
    
    def _update_yi_remaining(self):
        '''
        
        Update yi_remaining in light of the state of the boxes associated
        with this prim instance.
        
        '''
        
        # set the indices
        logical = np.ones(self.yi.shape[0],dtype=np.bool )
        for box in self._boxes:
            logical[box.yi] = False
        self.yi_remaining = self.yi[logical]
    
    def _peel(self, box):
        '''
        
        Executes the peeling phase of the PRIM algorithm. Delegates peeling
        to data type specific helper methods.

        '''
    
        mass_old = box.yi.shape[0]/self.n

        x = self.x[box.yi]
       
        #identify all possible peels
        possible_peels = []
        for entry in x.dtype.descr:
            u = entry[0]
            dtype = x.dtype.fields.get(u)[0].name
            peels = self._peels[dtype](self, box, u, x)
            [possible_peels.append(entry) for entry in peels] 
        if not possible_peels:
            # there is no peel identified, so return box
            return box

        # determine the scores for each peel in order
        # to identify the next candidate box
        scores = []
        for entry in possible_peels:
            i, box_lim = entry
            obj = self.obj_func(self, self.y[box.yi],  self.y[i])
            non_res_dim = len(x.dtype.descr)-\
                          sdutil._determine_nr_restricted_dims(box_lim, 
                                                              self.box_init)
            score = (obj, non_res_dim, box_lim, i)
            scores.append(score)

        scores.sort(key=itemgetter(0,1), reverse=True)
        entry = scores[0]
        
        obj_score = entry[0]
        box_new, indices = entry[2:]
        
        mass_new = self.y[indices].shape[0]/self.n
       
        if (mass_new >= self.mass_min) &\
           (mass_new < mass_old)&\
           (obj_score>0):
            box.update(box_new, indices)
            return self._peel(box)
        else:
            #else return received box
            return box
    
    
    def _real_peel(self, box, u, x):
        '''
        
        returns two candidate new boxes, peel along upper and lower dimension
        
        Parameters
        ----------
        box : a PrimBox instance
        u : str
            the uncertainty for which to peel
        
        Returns
        -------
        tuple
            two box lims and the associated indices
        
        '''

        peels = []
        for direction in ['upper', 'lower']:
            
            if not np.any(np.isnan(x[u])):
                peel_alpha = self.peel_alpha
            
                i=0
                if direction=='upper':
                    peel_alpha = 1-self.peel_alpha
                    i=1
                
                box_peel = get_quantile(x[u], peel_alpha)
                if direction=='lower':
                    logical = x[u] >= box_peel
                    indices = box.yi[logical]
                if direction=='upper':
                    logical = x[u] <= box_peel
                    indices = box.yi[logical]
                temp_box = copy.deepcopy(box.box_lims[-1])
                temp_box[u][i] = box_peel
                peels.append((indices, temp_box))
            else:
                return []
    
        return peels
    
    def _discrete_peel(self, box, u, x):
        '''
        
        returns two candidate new boxes, peel along upper and lower dimension
        
        Parameters
        ----------
        box : a PrimBox instance
        u : str
            the uncertainty for which to peel
        
        Returns
        -------
        tuple
            two box lims and the associated indices
        
        '''
        peels = []
        for direction in ['upper', 'lower']:
            peel_alpha = self.peel_alpha
        
            i=0
            if direction=='upper':
                peel_alpha = 1-self.peel_alpha
                i=1
            
            box_peel = get_quantile(x[u], peel_alpha)
            box_peel = int(box_peel)

            # determine logical associated with peel value            
            if direction=='lower':
                if box_peel == box.box_lims[-1][u][i]:
                    logical = (x[u] > box.box_lims[-1][u][i]) &\
                              (x[u] <= box.box_lims[-1][u][i+1])
                else:
                    logical = (x[u] >= box_peel) &\
                              (x[u] <= box.box_lims[-1][u][i+1])
            if direction=='upper':
                if box_peel == box.box_lims[-1][u][i]:
                    logical = (x[u] < box.box_lims[-1][u][i]) &\
                              (x[u] >= box.box_lims[-1][u][i-1])
                else:
                    logical = (x[u] <= box_peel) &\
                              (x[u] >= box.box_lims[-1][u][i-1])

            # determine value of new limit given logical
            if x[logical].shape[0] == 0:
                if direction == 'upper':
                    new_limit = np.max(x[u])
                else:
                    new_limit = np.min(x[u])
            else:
                if direction =='upper':
                    new_limit = np.max(x[u][logical])
                else:
                    new_limit = np.min(x[u][logical])            
            
            indices= box.yi[logical] 
            temp_box = copy.deepcopy(box.box_lims[-1])
            temp_box[u][i] = new_limit
            peels.append((indices, temp_box))
    
        return peels
    
    def _categorical_peel(self, box, u, x):
        '''
        
        returns candidate new boxes for each possible removal of a single 
        category. So. if the box[u] is a categorical variable with 4 
        categories, this method will return 4 boxes. 
        
        Parameters
        ----------
        box : a PrimBox instance
        u : str
            the uncertainty for which to peel
        
        Returns
        -------
        tuple
            a list of box lims and the associated indices
        
        '''
        entries = box.box_lims[-1][u][0]
        
        if len(entries) > 1:
            peels = []
            for entry in entries:
                temp_box = np.copy(box.box_lims[-1])
                peel = copy.deepcopy(entries)
                peel.discard(entry)
                temp_box[u][:] = peel
                
                if type(list(entries)[0]) not in (str, float, 
                                                  int):
                    bools = []                
                    for element in list(x[u]):
                        if element != entry:
                            bools.append(True)
                        else:
                            bools.append(False)
                    logical = np.asarray(bools, dtype=bool)
                else:
                    logical = x[u] != entry
                indices = box.yi[logical]
                peels.append((indices,  temp_box))
            return peels
        else:
            # no peels possible, return empty list
            return []

    def _paste(self, box):
        ''' Executes the pasting phase of the PRIM. Delegates pasting to data 
        type specific helper methods.'''
        
        x = self.x[self.yi_remaining]
        
        mass_old = box.yi.shape[0]/self.n
        
        res_dim = sdutil._determine_restricted_dims(box.box_lims[-1],
                                                    self.box_init)
        
        possible_pastes = []
        for u in res_dim:
            debug("pasting "+u)
            dtype = self.x.dtype.fields.get(u)[0].name
            pastes = self._pastes[dtype](self, box, u)
            [possible_pastes.append(entry) for entry in pastes] 
        if not possible_pastes:
            # there is no peel identified, so return box
            return box
    
        # determine the scores for each peel in order
        # to identify the next candidate box
        scores = []
        for entry in possible_pastes:
            i, box_lim = entry
            obj = self.obj_func(self, self.y[box.yi],  self.y[i])
            non_res_dim = len(x.dtype.descr)-\
                          sdutil._determine_nr_restricted_dims(box_lim,
                                                              self.box_init)
            score = (obj, non_res_dim, box_lim, i)
            scores.append(score)

        scores.sort(key=itemgetter(0,1), reverse=True)
        entry = scores[0]
        obj, _, box_new, indices = entry
        mass_new = self.y[indices].shape[0]/self.n
        
        mean_old = np.mean(self.y[box.yi])
        mean_new = np.mean(self.y[indices])
        
        if (mass_new >= self.mass_min) &\
           (mass_new > mass_old) &\
           (obj>0) &\
           (mean_new>mean_old):
            box.update(box_new, indices)
            return self._paste(box)
        else:
            #else return received box
            return box

    def _real_paste(self, box, u):
        '''
        
        returns two candidate new boxes, pasted along upper and lower 
        dimension
        
        Parameters
        ----------
        box : a PrimBox instance
        u : str
            the uncertainty for which to peel
        
        Returns
        -------
        tuple
            two box lims and the associated indices
       
        '''

        pastes = []
        for i, direction in enumerate(['lower', 'upper']):
            box_paste = np.copy(box.box_lims[-1])
            paste_box = np.copy(box.box_lims[-1]) # box containing data candidate for pasting
            
            if direction == 'upper':
                paste_box[u][0] = paste_box[u][1]
                paste_box[u][1] = self.box_init[u][1]
                indices = sdutil._in_box(self.x[self.yi_remaining], paste_box)
                data = self.x[self.yi_remaining][indices][u]
                
                paste_value = self.box_init[u][i]
                if data.shape[0] > 0:
                    paste_value = get_quantile(data, self.paste_alpha)
                    
                assert paste_value >= box.box_lims[-1][u][i]
                    
            elif direction == 'lower':
                paste_box[u][0] = self.box_init[u][0]
                paste_box[u][1] = box_paste[u][0]
                
                indices = sdutil._in_box(self.x[self.yi_remaining], paste_box)
                data = self.x[self.yi_remaining][indices][u]
                
                paste_value = self.box_init[u][i]
                if data.shape[0] > 0:
                    paste_value = get_quantile(data, 1-self.paste_alpha)
           
                if not paste_value <= box.box_lims[-1][u][i]:
                    print("{}, {}".format(paste_value, box.box_lims[-1][u][i]))
            
            
            dtype = box_paste.dtype.fields[u][0]
            if dtype==np.int32:
                paste_value = np.int(paste_value)
            
            box_paste[u][i] = paste_value
            indices = sdutil._in_box(self.x[self.yi_remaining], box_paste)
            indices = self.yi_remaining[indices]
            
            pastes.append((indices, box_paste))
    
        return pastes        
            
    def _categorical_paste(self, box, u):
        '''
        
        Return a list of pastes, equal to the number of classes currently
        not on the box lim. 
        
        Parameters
        ----------
        box : a PrimBox instance
        u : str
            the uncertainty for which to peel
        
        Returns
        -------
        tuple
            a list of box lims and the associated indices
        
        
        '''
        box_lim = box.box_lims[-1]
        
        c_in_b = box_lim[u][0]
        c_t = self.box_init[u][0]
        
        if len(c_in_b) < len(c_t):
            pastes = []
            possible_cs = c_t - c_in_b
            for entry in possible_cs:
                paste = copy.deepcopy(c_in_b)
                paste.add(entry)

                box_paste = np.copy(box_lim)
                box_paste[u][:] = paste
                
                indices = sdutil._in_box(self.x[self.yi_remaining], box_paste)
                indices = self.yi_remaining[indices]
                pastes.append((indices, box_paste))
            return pastes
        else:
            # no pastes possible, return empty list
            return []
    
    def _lenient1_obj_func(self, y_old, y_new):
        r'''
        the default objective function used by prim, instead of the original
        objective function, This function can cope with continuous, integer, 
        and categorical uncertainties. The basic idea is that the gain in mean
        is divided by the loss in mass. 
        
        .. math::
            
            obj = \frac
                 {\text{ave} [y_{i}\mid x_{i}\in{B-b}] - \text{ave} [y\mid x\in{B}]}
                 {|n(y_{i})-n(y)|}
        
        where :math:`B-b` is the set of candidate new boxes, :math:`B` 
        the old box and :math:`y` are the y values belonging to the old 
        box. :math:`n(y_{i})` and :math:`n(y)` are the cardinality of 
        :math:`y_{i}` and :math:`y` respectively. So, this objective 
        function looks for the difference between  the mean of the old 
        box and the new box, divided by the change in the  number of 
        data points in the box. This objective function offsets a problem 
        in case of categorical data where the normal objective function often 
        results in boxes mainly based on the categorical data.  
        
        '''
        mean_old = np.mean(y_old)
        
        if y_new.shape[0]>0:
            mean_new = np.mean(y_new)
        else:
            mean_new = 0
            
        obj = 0
        if mean_old != mean_new:
            if y_old.shape[0] > y_new.shape[0]:
                obj = (mean_new-mean_old)/(y_old.shape[0]-y_new.shape[0])
            elif y_old.shape[0] < y_new.shape[0]:
                obj = (mean_new-mean_old)/(y_new.shape[0]-y_old.shape[0])
            else:
                raise PrimException('''mean is different {} vs {}, while shape is the same,
                                       this cannot be the case'''.format(mean_old, mean_new))
        return obj
    
    def _lenient2_obj_func(self, y_old, y_new):
        '''
        
        friedman and fisher 14.6
        
        
        '''
        mean_old = np.mean(y_old)
        
        if y_new.shape[0]>0:
            mean_new = np.mean(y_new)
        else:
            mean_new = 0
            
        obj = 0
        if mean_old != mean_new:
            if y_old.shape==y_new.shape:
                raise PrimException('''mean is different {} vs {}, while shape is the same,
                                       this cannot be the case'''.format(mean_old, mean_new))
            
            change_mean = mean_new - mean_old
            change_mass = abs(y_old.shape[0]-y_new.shape[0])
            mass_new = y_new.shape[0]
                
            obj = mass_new * change_mean / change_mass
                
        return obj
    
    def _original_obj_func(self, y_old, y_new):
        ''' The original objective function: the mean of the data inside the 
        box'''
        
        if y_new.shape[0]>0:
            return np.mean(y_new)
        else:
            return -1    

    def _assert_dtypes(self, keys, dtypes):
        '''
        helper fucntion that checks whether none of the provided keys has
        a dtype object as value.
        '''
        
        for key in keys:
            if dtypes[key][0] == np.dtype(object):
                raise EMAError("%s has dtype object and can thus not be rotated" %key)
        return True

    def _rotate_subset(self, value, orig_experiments, logical): 
        '''
        rotate a subset
        
        Parameters
        ----------
        value : list of str
        orig_experiment : numpy structured array
        logical : boolean array
        
        '''
        list_dtypes = [(name, "<f8") for name in value]
        
        #cast everything to float
        drop_names = set(rf.get_names(orig_experiments.dtype)) - set(value)
        orig_subset = rf.drop_fields(orig_experiments, drop_names, 
                                               asrecarray=True)
        subset_experiments = orig_subset.astype(list_dtypes).view('<f8').reshape(orig_experiments.shape[0], len(value))
 
        #normalize the data
        mean = np.mean(subset_experiments,axis=0)
        std = np.std(subset_experiments, axis=0)
        std[std==0] = 1 #in order to avoid a devision by zero
        subset_experiments = (subset_experiments - mean)/std
        
        #get the experiments of interest
        experiments_of_interest = subset_experiments[logical]
        
        #determine the rotation
        rotation_matrix =  self._determine_rotation(experiments_of_interest)
        
        #apply the rotation
        subset_experiments = np.dot(subset_experiments,rotation_matrix)
        return rotation_matrix, subset_experiments

    def _determine_rotation(self, experiments):
        '''
        Determine the rotation for the specified experiments
        
        '''
        covariance = np.cov(experiments.T)
        
        eigen_vals, eigen_vectors = np.linalg.eig(covariance)
    
        indices = np.argsort(eigen_vals)
        indices = indices[::-1]
        eigen_vectors = eigen_vectors[:,indices]
        eigen_vals = eigen_vals[indices]
        
        #make the eigen vectors unit length
        for i in range(eigen_vectors.shape[1]):
            eigen_vectors[:,i] / np.linalg.norm(eigen_vectors[:,i]) * np.sqrt(eigen_vals[i])
            
        return eigen_vectors

    _peels = {'object': _categorical_peel,
              'int64': _discrete_peel,
               'int32': _discrete_peel,
               'float64': _real_peel}

    _pastes = {'object': _categorical_paste,
               'int32': _real_paste,
               'int64': _real_paste,
               'float64': _real_paste}

    # dict with the various objective functions available
    _obj_functions = {LENIENT2 : _lenient2_obj_func,
                      LENIENT1 : _lenient1_obj_func,
                      ORIGINAL: _original_obj_func}    


class MultiBoxesPrim(Prim):
    '''Modification of PRIM's handling of multiple boxes, based on the
    suggestion of Guivarch et al (2016) doi:10.1016/j.envsoft.2016.03.006
    
    
    TODO:: we need a better name for this
    
    '''
    
    def _update_yi_remaining(self):
        '''
        
        Update yi_remaining in light of the state of the boxes associated
        with this prim instance.
        
        '''
        # set the indices
        for box in self._boxes:
            self.y[box.yi] = 0
        print(np.sum(self.y))
        
        self.yi_remaining = self.yi



'''
Scenario discovery utilities used by both :mod:`cart` and :mod:`prim`
'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

import abc

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import numpy.lib.recfunctions as recfunctions
import pandas as pd

from .plotting_util import COLOR_LIST
from ..util import ema_logging
import itertools


COLOR_LIST = itertools.cycle(COLOR_LIST)

# Created on May 24, 2015
# 
# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

REGRESSION = 'regression'
'''constant indicating regression mode'''

BINARY = 'binary'
'''constant indicating binary classification mode. This is the most
common used mode in scenario discovery'''

CLASSIFICATION = 'classification'
'''constant indicating classification mode'''


def _get_sorted_box_lims(boxes, box_init):
    '''Sort the uncertainties for each box in boxes based on a normalization
    given box_init. Unrestricted dimensions are dropped. The sorting is based
    on the normalization of the first box in boxes. 
    
    Parameters
    ----------
    boxes : list of numpy structured arrays
    box_init : numpy structured array
    
    Returns
    -------
    tuple 
        with the sorted boxes, and the list of restricted uncertainties
    
    '''
        
    # determine the uncertainties that are being restricted
    # in one or more boxes
    uncs = set()
    for box in boxes:
        us  = _determine_restricted_dims(box, box_init).tolist()
        uncs = uncs.union(us)
    uncs = np.asarray(list(uncs))

    # normalize the range for the first box
    box_lim = boxes[0]
    nbl = _normalize(box_lim, box_init, uncs)
    box_size = nbl[:,1]-nbl[:,0]
    
    # sort the uncertainties based on the normalized size of the 
    # restricted dimensions
    uncs = uncs[np.argsort(box_size)]
    box_lims = [box for box in boxes]
    
    return box_lims, uncs.tolist()


def _make_box(x):
    '''
    Make a box that encompasses all the data
    
    Parameters
    ----------
    x : structured numpy array
    
    
    '''
    
    box = np.zeros((2, ), x.dtype)
    
    names = recfunctions.get_names(x.dtype)
    
    for name in names:
        dtype = x.dtype.fields.get(name)[0] 
        mask = np.ma.getmaskarray(x[name])
        values = x[name][mask==False]
        
        if dtype == 'object':
            try:
                values = set(values)
            except TypeError as e:
                ema_logging.warning("{} has unhashable values".format(name))
                raise e
            else:
                box[name][:] = values
        else:
            box[name][0] = np.min(values, axis=0) 
            box[name][1] = np.max(values, axis=0)    
    return box  


def _normalize(box_lim, box_init, uncertainties):
    '''Normalize the given box lim to the unit interval derived
    from box init for the specified uncertainties.
    
    Categorical uncertainties are normalized based on fractionated. So
    value specifies the fraction of categories in the box_lim. 
    
    Parameters
    ----------
    box_lim : a numpy structured array.
    box_init :  a numpy structured array.
    uncertainties : list of strings
                    valid names of columns that exist in both structured 
                    arrays.

    Returns
    -------
    ndarray
        a numpy array of the shape (2, len(uncertainties) with the box limits.
    
    
    '''
    
    # normalize the range for the first box
    norm_box_lim = np.zeros((len(uncertainties), box_lim.shape[0]))
    
    for i, u in enumerate(uncertainties):
        dtype = box_lim.dtype.fields[u][0]
        if dtype ==np.dtype(object):
            nu = len(box_lim[u][0])/ len(box_init[u][0])
            nl = 0
        else:
            lower, upper = box_lim[u]
            dif = (box_init[u][1]-box_init[u][0])
            a = 1/dif
            b = -1 * box_init[u][0] / dif
            nl = a * lower + b
            nu = a * upper + b
        norm_box_lim[i, :] = nl, nu
    return norm_box_lim


def _determine_restricted_dims(box_lims, box_init):
    '''
    
    determine which dimensions of the given box are restricted compared 
    to compared to the initial box that contains all the data
    
    Parameters
    ----------
    box_lims : structured numpy array
               a specific box limit
    box_init : structured numpy array
               the initial box containing all data points
    
    '''

    logical = _compare(box_init, box_lims)
    u = np.asarray(recfunctions.get_names(box_lims.dtype), 
                   dtype=object)
    dims = u[logical==False]
    return dims


def _determine_nr_restricted_dims(box_lims, box_init):
    '''
    
    determine the number of restriced dimensions of a box given
    compared to the inital box that contains all the data
    
    Parameters
    ----------
    box_lims : structured numpy array
               a specific box limit
    box_init : structured numpy array
               the initial box containing all data points
    
    '''

    return _determine_restricted_dims(box_lims, box_init).shape[0]

def _compare(a, b):
    '''compare two boxes, for each dimension return True if the
    same and false otherwise'''
    dtypesDesc = a.dtype.descr
    logical = np.ones((len(dtypesDesc,)), dtype=np.bool)
    for i, entry in enumerate(dtypesDesc):
        name = entry[0]
        logical[i] = logical[i] &\
                    (a[name][0] == b[name][0]) &\
                    (a[name][1] == b[name][1])
    return logical


def _setup_figure(uncs):
    '''
    
    helper function for creating the basic layout for the figures that
    show the box lims.
    
    '''
    nr_unc = len(uncs)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    
    # create the shaded grey background
    rect = mpl.patches.Rectangle((0, -0.5), 1, nr_unc+1.5,
                                 alpha=0.25,  
                                 facecolor="#C0C0C0",
                                 edgecolor="#C0C0C0")
    ax.add_patch(rect)
    ax.set_xlim(xmin=-0.2, xmax=1.2)
    ax.set_ylim(ymin= -0.5, ymax=nr_unc-0.5)
    ax.yaxis.set_ticks([y for y in range(nr_unc)])
    ax.xaxis.set_ticks([0, 0.25, 0.5, 0.75, 1])
    ax.set_yticklabels(uncs[::-1]) 
    return fig, ax


def _in_box(x, boxlim):
    '''
     
    returns the indices of the data points that are within the 
    box_lims.
    
    Parameters
    ----------
    x : numpy structured array
    boxlim : numpy structured array
    
    
    Returns
    -------
    ndarray
        valid numpy indices on x
    
    '''
    logical = np.ones(x.shape[0], dtype=np.bool)
    
    dims = recfunctions.get_names(boxlim.dtype)

    for name in dims:
        value = x.dtype.fields.get(name)[0]
        
        if value == 'object':
            entries = boxlim[name][0]
            l = np.ones( (x.shape[0], len(entries)), dtype=np.bool)
            for i,entry in enumerate(entries):
                if type(list(entries)[0]) not in (str, float, int):
                    bools = []                
                    for element in list(x[name]):
                        if element == entry:
                            bools.append(True)
                        else:
                            bools.append(False)
                    l[:, i] = np.asarray(bools, dtype=bool)
                else:
                    l[:, i] = x[name] == entry
            l = np.any(l, axis=1)
            logical = logical & l
        else:
            logical = logical & (boxlim[name][0] <= x[name] )&\
                                        (x[name] <= boxlim[name][1])                
    
    indices = np.where(logical==True)
    
    assert len(indices)==1
    indices = indices[0]
    
    return indices


class OutputFormatterMixin(object):
    __metaclass__ = abc.ABCMeta
    
    @abc.abstractproperty
    def boxes(self):
        '''Property for getting a list of box limits'''
        
        raise NotImplementedError
    
    @abc.abstractproperty
    def stats(self):
        '''property for getting a list of dicts containing the statistics
        for each box'''
        
        raise NotImplementedError
    
    def boxes_to_dataframe(self):
        '''convert boxes to pandas dataframe'''
        
        boxes = self.boxes
            
        # determine the restricted dimensions
        # print only the restricted dimension
        box_lims, uncs = _get_sorted_box_lims(boxes, _make_box(self.x))
        nr_boxes = len(boxes)
        dtype = float
        index = ["box {}".format(i+1) for i in range(nr_boxes)]
        for value in box_lims[0].dtype.fields.values():
            if value[0] == object:
                dtype = object
                break
                
        columns = pd.MultiIndex.from_product([index,
                                              ['min', 'max',]])
        df_boxes = pd.DataFrame(np.zeros((len(uncs), nr_boxes*2)),
                               index=uncs,
                               dtype=dtype,
                               columns=columns)

        for i, box in enumerate(box_lims):
            for unc in uncs:
                values = box[unc][:]
                values = pd.Series(values, 
                                   index=['min','max'])
                df_boxes.ix[unc][index[i]] = values   
        return df_boxes 
    
    def stats_to_dataframe(self):
        '''convert stats to pandas dataframe'''
        
        stats = self.stats
        
        index = pd.Index(['box {}'.format(i+1) for i in range(len(stats))])
        
        return pd.DataFrame(stats, index=index)
    
    def display_boxes(self, together=False):
        '''display boxes
        
        Parameters
        ----------
        together : bool, otional
        
        '''
        box_init = _make_box(self.x)
        box_lims, uncs = _get_sorted_box_lims(self.boxes, box_init)

        # normalize the box lims
        # we don't need to show the last box, for this is the 
        # box_init, which is visualized by a grey area in this
        # plot.
        norm_box_lims =  [_normalize(box_lim, box_init, uncs) for 
                                        box_lim in box_lims[0:-1]]
                        
        if together:
            fig, ax = _setup_figure(uncs)
            color = next(COLOR_LIST)
            
            for i, u in enumerate(uncs):
                # we want to have the most restricted dimension
                # at the top of the figure
                xi = len(uncs) - i - 1
                
                for j, norm_box_lim in enumerate(norm_box_lims):
                    self._plot_unc(box_init, xi, i, j, norm_box_lim,
                                   box_lims[j], u, ax, color)
           
            plt.tight_layout()
            return fig
        else:
            figs = []
            color = next(COLOR_LIST)
            for j, norm_box_lim in enumerate(norm_box_lims):
                fig, ax = _setup_figure(uncs)
                figs.append(fig)
                for i, u in enumerate(uncs):
                    xi = len(uncs) - i - 1
                    self._plot_unc(box_init, xi, i, j, norm_box_lim, 
                                   box_lims[j], u, ax, color)
        
                plt.tight_layout()
            return figs
    
    @staticmethod  
    def _plot_unc(box_init, xi, i, j, norm_box_lim, box_lim, u, ax, color):
        '''
        
        Parameters:
        ----------
        xi : int 
             the row at which to plot
        i : int
            the index of the uncertainty being plotted
        j : int
            the index of the box being plotted
        u : string
            the uncertainty being plotted:
        ax : axes instance
             the ax on which to plot
        
        '''

        dtype = box_init[u].dtype
            
        y = xi-j*0.1
        
        if dtype == object:
            elements = sorted(list(box_init[u][0]))
            max_value = (len(elements)-1)
            box_lim = box_lim[u][0]
            x = [elements.index(entry) for entry in 
                 box_lim]
            x = [entry/max_value for entry in x]
            y = [y] * len(x)
            
            ax.scatter(x,y,  edgecolor=color,
                       facecolor=color)
            
        else:
            ax.plot(norm_box_lim[i], (y, y),
                    c=color)
        
        


'''
Created on 11 aug. 2011

.. codeauthor:: wauping <w.auping (at) student (dot) tudelft (dot) nl>
                jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>


To be able to debug the Vensim model, a few steps are needed:
    
    1.  The case that gave a bug, needs to be saved in a text  file. The entire 
        case description should be on a single line.
    2.  Reform and clean your model ( In the Vensim menu: Model, Reform and 
        Clean). Choose    
        
         * Equation Order: Alphabetical by group (not really necessary)
         * Equation Format: Terse
        
    3.  Save your model as text (File, Save as..., Save as Type: Text Format 
        Models
    4.  Run this script
    5.  If the print in the end is not set([]), but set([array]), the array 
        gives the values that where not found and changed
    5.  Run your new model (for example 'new text.mdl')
    6.  Vensim tells you about your critical mistake

'''
from __future__ import (division, print_function, unicode_literals)

fileSpecifyingError = ""

pathToExistingModel = r"C:\workspace\EMA-workbench\models\salinization\Verzilting_aanpassingen incorrect.mdl"
pathToNewModel  = r"C:\workspace\EMA-workbench\models\salinization\Verzilting_aanpassingen correct.mdl"
newModel = open(pathToNewModel, 'w')   

#line = open(fileSpecifyingError).read()

line = 'rainfall : 0.154705633188; adaptation time from non irrigated agriculture : 0.915157119079; salt effect multiplier : 1.11965969891; adaptation time to non irrigated agriculture : 0.48434342934; adaptation time to irrigated agriculture : 0.330990830832; water shortage multiplier : 0.984356102036; delay time salt seepage : 6.0; adaptation time : 6.90258192256; births multiplier : 1.14344734715; diffusion lookup : [(0, 8.0), (10, 8.0), (20, 8.0), (30, 8.0), (40, 7.9999999999999005), (50, 4.0), (60, 9.982194802803703e-14), (70, 1.2455526635140464e-27), (80, 1.5541686655435471e-41), (90, 1.9392517969836692e-55)]; salinity effect multiplier : 1.10500381093; technological developments in irrigation : 0.0117979353255; adaptation time from irrigated agriculture : 1.58060947607; food shortage multiplier : 0.955325345996; deaths multiplier : 0.875605669911; '

# we assume the case specification was copied from the logger
splitOne = line.split(';')                              
variable = {}
for n in range(len(splitOne)-1):
    splitTwo = splitOne[n].split(':')
    variableElement=splitTwo[0]
    # Delete the spaces and other rubish on the sides of the variable name
    variableElement=variableElement.lstrip()            
    variableElement=variableElement.lstrip("'")
    variableElement=variableElement.rstrip()
    variableElement=variableElement.rstrip("'")
    print(variableElement)
    valueElement=splitTwo[1]
    valueElement=valueElement.lstrip()
    valueElement=valueElement.rstrip()
    variable[variableElement] = valueElement
print(variable)

# This generates a new (text-formatted) model
changeNextLine = False
settedValues = []
for line  in open(pathToExistingModel):    
    if line.find("=")!= -1:
        elements = line.split("=")
        value = elements[0]
        value = value.strip()
        if value in variable:
            elements[1] = variable.get(value)
            line = elements[0] + " = " + elements[1]
            settedValues.append(value)

    newModel.write(line) 
notSet = set(variable.keys())-set(settedValues)         
print(notSet)


'''
An example of the lake problem using the ema workbench. 

The model itself is adapted from the Rhodium example by Dave Hadka,
see https://gist.github.com/dhadka/a8d7095c98130d8f73bc

'''
from __future__ import (unicode_literals, print_function, absolute_import,
                        division)

import math

import numpy as np
from scipy.optimize import brentq

from ema_workbench import (Model, RealParameter, ScalarOutcome, Constant,
                           ema_logging, MultiprocessingEvaluator)
from ema_workbench.em_framework.evaluators import MC, SequentialEvaluator
from ema_workbench.em_framework.parameters import CategoricalParameter


def lake_problem(
         b = 0.42,          # decay rate for P in lake (0.42 = irreversible)
         q = 2.0,           # recycling exponent
         mean = 0.02,       # mean of natural inflows
         stdev = 0.001,     # future utility discount rate
         delta = 0.98,      # standard deviation of natural inflows
         alpha = 0.4,       # utility from pollution
         nsamples = 100,    # Monte Carlo sampling of natural inflows
         **kwargs):   
    try:
        decisions = [kwargs[str(i)] for i in range(100)]
    except KeyError:
        decisions = [0, ] * 100
    
    Pcrit = brentq(lambda x: x**q/(1+x**q) - b*x, 0.01, 1.5)
    nvars = len(decisions)
    X = np.zeros((nvars,))
    average_daily_P = np.zeros((nvars,))
    decisions = np.array(decisions)
    reliability = 0.0

    for _ in range(nsamples):
        X[0] = 0.0
        
        natural_inflows = np.random.lognormal(
                math.log(mean**2 / math.sqrt(stdev**2 + mean**2)),
                math.sqrt(math.log(1.0 + stdev**2 / mean**2)),
                size = nvars)
        
        for t in range(1,nvars):
            X[t] = (1-b)*X[t-1] + X[t-1]**q/(1+X[t-1]**q) + decisions[t-1] + natural_inflows[t-1]
            average_daily_P[t] += X[t]/float(nsamples)
    
        reliability += np.sum(X < Pcrit)/float(nsamples*nvars)
      
    max_P = np.max(average_daily_P)
    utility = np.sum(alpha*decisions*np.power(delta,np.arange(nvars)))
    inertia = np.sum(np.diff(decisions) > -0.02)/float(nvars-1)

    return max_P, utility, inertia, reliability


if __name__ == '__main__':
    ema_logging.log_to_stderr(ema_logging.INFO)

    # instantiate the model
    lake_model = Model('lakeproblem', function=lake_problem)
    lake_model.time_horizon = 100
    
    #specify uncertainties
    lake_model.uncertainties = [RealParameter('b', 0.1, 0.45),
                                RealParameter('q', 2.0, 4.5),
                                RealParameter('mean', 0.01, 0.05),
                                RealParameter('stdev', 0.001, 0.005),
                                RealParameter('delta', 0.93, 0.99)]
    
    # set levers, one for each time step
    lake_model.levers = [RealParameter(str(i), 0, 0.1) for i in 
                         range(lake_model.time_horizon)]
    
    #specify outcomes 
    lake_model.outcomes = [ScalarOutcome('max_P',),
                           ScalarOutcome('utility'),
                           ScalarOutcome('inertia'),
                           ScalarOutcome('reliability')]
    
    # override some of the defaults of the model
    lake_model.constants = [Constant('alpha', 0.41),
                            Constant('nsamples', 150)]
        
    # generate some random policies by sampling over levers
    n_scenarios = 100
    n_policies = 4
    
     
    with MultiprocessingEvaluator(lake_model) as evaluator:
        res = evaluator.perform_experiments(n_scenarios, n_policies,
                                            levers_sampling=MC)


'''
Created on Feb 13, 2014

This example demonstrates the use of PRIM. The dataset was generated
using the world container model (Tavasszy et al 2011; http://dx.doi.org/10.1016/j.jtrangeo.2011.05.005)

@author: jhkwakkel
'''
import numpy as np
import matplotlib.pyplot as plt

from ema_workbench import ema_logging, load_results
from ema_workbench.analysis import prim

ema_logging.log_to_stderr(ema_logging.INFO);

default_flow = 2.178849944502783e7

def classify(outcomes):
    ooi = 'throughput Rotterdam'
    outcome = outcomes[ooi]
    outcome = outcome/default_flow
    
    classes = np.zeros(outcome.shape[0])
    classes[outcome<1] = 1
    return classes

fn = r'./data/5000 runs WCM.tar.gz'
results = load_results(fn)


prim_obj = prim.setup_prim(results, classify, mass_min=0.05, threshold=0.75)

# let's find a first box
box1 = prim_obj.find_box()

# let's analyze the peeling trajectory
box1.show_ppt()
box1.show_tradeoff()

box1.write_ppt_to_stdout()

# based on the peeling trajectory, we pick entry number 44
box1.select(44)

# show the resulting box
prim_obj.show_boxes()
prim_obj.write_boxes_to_stdout()

plt.show()


'''


'''
from __future__ import (unicode_literals, print_function, absolute_import,
                        division)

from collections import defaultdict
import io
import logging
import multiprocessing
import os
import sys
import threading
import time
import shutil
import traceback

from ..util import ema_logging
from .experiment_runner import ExperimentRunner
from .util import NamedObjectMap
from .model import AbstractModel

# Created on 22 Feb 2017
#
# .. codeauthor::jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

__all__ = []


def initializer(*args):
    '''initializer for a worker process
    
    Parameters
    ----------
    models : list of AbstractModel instances
    
    
    This function initializes the worker. This entails
    * initializing the experiment runner
    * setting up the working directory
    * setting up the logging
    '''
    global experiment_runner, current_process
    
    current_process = multiprocessing.current_process()
    models, queue, log_level, root_dir = args
    
    # setup the experiment runner
    msis = NamedObjectMap(AbstractModel)
    msis.extend(models)
    experiment_runner = ExperimentRunner(msis)
    
    # setup the logging 
    setup_logging(queue, log_level)
    
    # setup the working directories
    # make a root temp
    # copy each model directory
    tmpdir = setup_working_directories(models, root_dir)
        
    # register a cleanup finalizer function
    # remove the root temp
    if tmpdir:
        multiprocessing.util.Finalize(None, finalizer, args=(os.path.abspath(tmpdir), ), 
                                  exitpriority=10)  # @UndefinedVariable


def finalizer(tmpdir):
    '''cleanup'''
    global experiment_runner
    ema_logging.info("finalizing")
    
    experiment_runner.cleanup()
    del experiment_runner
    
    time.sleep(1)
    
    if tmpdir:
        try:
            shutil.rmtree(tmpdir)
        except OSError:
            pass
    

def setup_logging(queue, log_level):
    '''helper function for enabling logging from the workers to the main 
    process
    
    Parameters
    ----------
    queue : multiprocessing.Queue instance
    log_level : int
    
    '''
    
    # create a logger
    logger = logging.getLogger(ema_logging.LOGGER_NAME+'.subprocess')
    ema_logging._logger = logger
    logger.handlers = []

    # add the handler
    handler = SubProcessLogHandler(queue)
    handler.setFormatter(ema_logging.LOG_FORMAT)
    logger.addHandler(handler)
    
    # set the log_level
    logger.setLevel(log_level)


def setup_working_directories(models, root_dir):
    '''copies the working directory of each model to a process specific 
    temporary directory and update the working directory of the model
    
    Parameters
    ----------
    models : list
    root_dir : str
    
    '''
    
    # group models by working directory to avoid copying the same directory
    # multiple times
    wd_by_model = defaultdict(list)
    for model in models:
        try:
            wd = model.working_directory
        except AttributeError:
            pass
        else:
            wd_by_model[wd].append(model)
    
    # if the dict is not empty
    if wd_by_model:
        # make a directory with the process id as identifier
        tmpdir_name = "tmp{}".format(os.getpid())
        tmpdir = os.path.join(root_dir, tmpdir_name)
        os.mkdir(tmpdir)
        
        ema_logging.debug("setting up working directory: {}".format(tmpdir))
        
        for key, value in wd_by_model.items():
            # we need a sub directory in the process working directory
            # for each unique model working directory
            subdir = os.path.basename(os.path.normpath(key))
            new_wd = os.path.join(tmpdir, subdir)
            
            # the copy operation
            shutil.copytree(key, new_wd)
            
            for model in value:
                model.working_directory = new_wd
        return tmpdir
    else:
        return None
    

def worker(experiment):
    '''the worker function for executing an individual experiment
    
    Parameters
    ----------
    experiment : dict
    
    '''
    global experiment_runner
    return experiment_runner.run_experiment(experiment)


class SubProcessLogHandler(logging.Handler):
    """handler used by subprocesses

    It simply puts items on a Queue for the main process to log.

    adapted a bit using code found in same stack overflow thread 
    so that exceptions can be logged. Exception stack traces cannot be pickled
    so they cannot be put into the queue. Therefore they are formatted first 
    and then put into the queue as a normal message
    
    Found `online <http://stackoverflow.com/questions/641420/how-should-i-log-while-using-multiprocessing-in-python>`_

    """

    def __init__(self, queue):
        logging.Handler.__init__(self)
        self.queue = queue

    def emit(self, record):
        if record.exc_info:
            # can't pass exc_info across processes so just format now
            record.exc_text = self.formatException(record.exc_info)
            record.exc_info = None
        self.queue.put(record)

    def formatException(self, ei):
        sio = io.StringIO()
        traceback.print_exception(ei[0], ei[1], ei[2], None, sio)
        s = sio.getvalue()
        sio.close()
        if s[-1] == "\n":
            s = s[:-1]
        return s


class LogQueueReader(threading.Thread):
    """
    
    thread to write subprocesses log records to main process log

    This thread reads the records written by subprocesses and writes them to
    the handlers defined in the main process's handlers.
    
    found `online <http://stackoverflow.com/questions/641420/how-should-i-log-while-using-multiprocessing-in-python>`_
    
    
    TODO:: should be generalized with logwatcher used with ipyparallel

    """

    def __init__(self, queue):
        threading.Thread.__init__(self,name="log queue reader")
        self.queue = queue
        self.daemon = True

    def run(self):
        """
        read from the queue and write to the log handlers

        The logging documentation says logging is thread safe, so there
        shouldn't be contention between normal logging (from the main
        process) and this thread.

        Note that we're using the name of the original logger.

        """
        
        while True:
            try:
                record = self.queue.get()
                # get the logger for this record
                if record is None:
                    ema_logging.debug("none received")
                    break
                
                logger = logging.getLogger(record.name)
                logger.callHandlers(record)
            except (KeyboardInterrupt, SystemExit):
                raise
            except EOFError:
                break
            except TypeError:
                break
            except:
                traceback.print_exc(file=sys.stderr)


def result_handler(callback, experiment):
    '''handler for the results
    
    to link experiment and output, we use a functional programming
    trick.
    
    '''
    
    def my_actual_callback(result):
        callback(experiment, result)
    return my_actual_callback


def add_tasks(pool, experiments, callback):
    '''add experiments to pool
    
    '''
    
    results = []
    for e in experiments:
        
        # TODO:: code won't work on Python 3.4 or lower
        # error_callback only exists in 3.5 and up
        res = pool.apply_async(worker, [e], 
                                     callback=result_handler(callback, e))
        results.append(res)

    for res in results:
        res.wait()


'''

This module provides a base class that can be used to perform EMA on 
Excel models. It relies on `win32com <http://python.net/crew/mhammond/win32/Downloads.html>`_

'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

import os

import win32com.client  # @UnresolvedImport
from win32com.universal import com_error # @UnresolvedImport

from ..util import ema_logging, EMAError
from ..em_framework.model import FileModel
from ema_workbench.em_framework.model import SingleReplication
from ..util.ema_logging import method_logger

# Created on 19 sep. 2011
# 
# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

class BaseExcelModel(FileModel):
    '''
    
    Base class for connecting the EMA workbench to models in Excel. To 
    automate this connection as much as possible. This implementation relies
    on naming cells in Excel. These names can then be used here as names
    for the uncertainties and the outcomes. See e.g. `this site <http://spreadsheets.about.com/od/exceltips/qt/named_range.htm>`_
    for details on naming cells and sets of cells. 
    
    The provided implementation here does work with :mod:`parallel_ema`.
    
    '''
    
    com_warning_msg = "com error: no cell(s) named %s found"
    
    def __init__(self, name, wd=None, model_file=None):
        super(BaseExcelModel, self).__init__(name, wd=wd, model_file=model_file)
        #: Reference to the Excel application. This attribute is `None` until
        #: model_init has been invoked.
        self.xl = None
    
        #: Reference to the workbook. This attribute is `None` until
        #: model_init has been invoked.
        self.wb = None
    
        #: Name of the sheet on which one want to set values
        self.sheet = None

    @property
    def workbook(self):
        return self.model_file

    def model_init(self, policy):
        '''
        Method called to initialize the model.
        
        Parameters
        ----------
        policy : dict
                 policy to be run.
        kwargs : dict
                 keyword arguments to be used by model_intit. This
                 gives users to the ability to pass any additional 
                 arguments. 
        
        
        '''
        super(BaseExcelModel, self).model_init(policy)
        
        if not self.xl:
            try:
                ema_logging.debug("trying to start Excel")
                self.xl = win32com.client.Dispatch("Excel.Application")
                ema_logging.debug("Excel started") 
            except com_error as e:
                raise EMAError(str(e))
        
        # TODO for some strange reason, init is called for every replication
        if not self.wb:
            ema_logging.debug("trying to open workbook")
            wb = os.path.join(self.working_directory, self.workbook)
            self.wb = self.xl.Workbooks.Open(wb)
            ema_logging.debug("workbook opened")
            ema_logging.debug(self.working_directory)

    @method_logger
    def run_experiment(self, experiment):
        """
        Method for running an instantiated model structures. This 
        implementation assumes that the names of the uncertainties correspond
        to the name of the cells in Excel. See e.g. `this site <http://spreadsheets.about.com/od/exceltips/qt/named_range.htm>`_ 
        for details or use Google and search on 'named range'. One of the 
        requirements on the names is that they cannot contains spaces. 

        For the extraction of results, the same approach is used. That is, 
        this implementation assumes that the name of a :class:`~outcomes.Outcome`
        instance corresponds to the name of a cell, or set of cells.

        Parameters
        ----------
        scenario : Scenario instance
        policy : Policy instance
        
        
        """
#         super(ExcelModel, self).run_model(scenario, policy)
        
        #find right sheet
        try:
            sheet = self.wb.Sheets(self.sheet)
        except Exception :
            ema_logging.warning("com error: sheet not found")
            self.cleanup()
            raise
        
        #set values on sheet
        for key, value in experiment.items():
            try:
                sheet.Range(key).Value = value 
            except com_error:
                ema_logging.warning("com error: no cell(s) named %s found" % key,)

        #get results
        results = {}
        for variable in self.outcome_variables:
            try:
                output = sheet.Range(variable).Value
            except com_error:
                ema_logging.warning(self.com_warning_msg.format(variable))	
                continue
            results[variable] = output
            
        return results


    def cleanup(self):
        ''' cleaning up prior to finishing performing experiments. This will 
        close the workbook and close Excel'''
        
        ema_logging.debug("cleaning up")
        if self.wb:
            self.wb.Close(False)
            del self.wb
        if self.xl:
            self.xl.DisplayAlerts = False
            self.xl.Quit()
            del self.xl
        
        self.xl = None
        self.wb = None
        
class ExcelModel(SingleReplication, BaseExcelModel):
    pass


'''
Module for outcome classes

'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)
import abc
import six
import warnings

import pandas

from .util import Variable
from ema_workbench.util.ema_exceptions import EMAError


# Created on 24 mei 2011
# 
# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

__all__ = ['Outcome', 'ScalarOutcome', 'TimeSeriesOutcome']


def Outcome(name, time=False):
    if time:
        warnings.warn('Deprecated, use TimeSeriesOutcome instead')
        return ScalarOutcome(name)
    else:
        warnings.warn('Deprecated, use ScalarOutcome instead')
        return TimeSeriesOutcome(name)
    

class AbstractOutcome(Variable):
    '''
    Base Outcome class
    
    Parameters
    ----------
    name : str
           Name of the outcome.
    kind : {INFO, MINIMZE, MAXIMIZE}, optional
    variable_name : str, optional
                    if the name of the outcome in the underlying model
                    is different from the name of the outcome, you can 
                    supply the variable name as an optional argument,
                    if not provided, defaults to name
    function : callable, optional
               a callable to perform postprocessing on data retrieved from
               model
    
    Attributes
    ----------
    name : str
    kind : int
    
    '''
    __metaclass__ = abc.ABCMeta

    MINIMIZE = -1
    MAXIMIZE = 1
    INFO = 0
    
    def __init__(self, name, kind=INFO, variable_name=None, function=None):
        super(AbstractOutcome, self).__init__(name)
        
        if function is not None and not callable(function):
            raise ValueError('function must be a callable')
        if variable_name:
            if (not isinstance(variable_name, six.string_types)) and (not all(isinstance(elem, six.string_types) for elem in variable_name)):
                    raise ValueError('variable name must be a string or list of strings')
        
        
        self.kind = kind
        self.variable_name = variable_name
        self.function = function
    
    def process(self, values):
        if self.function:
            var_names = self.variable_name
            
            n_variables = len(var_names)
            try:
                n_values = len(values)
            except TypeError:
                len_val = None
                
            if (n_values==None) and (n_variables==1):
                return self.function(values)
            elif n_variables != n_values:
                raise ValueError(('number of variables is {}, '
                      'number of outputs is {}').format(n_variables, n_values))
            else:
                return self.function(*values)
        else: 
            if len(values)>1:
                raise EMAError(('more than one value returned without ' 
                                'processing function'))
            
            return values[0]
    
    def __eq__ (self, other):
        comparison = [all(hasattr(self, key) == hasattr(other, key) and
                          getattr(self, key) == getattr(other, key) for key 
                          in self.__dict__.keys())]
        comparison.append(self.__class__ == other.__class__)
        return all(comparison)
    
    def __repr__(self, *args, **kwargs):
        klass = self.__class__.__name__
        name = self.name
        
        rep = '{}(\'{}\''.format(klass, name)
        
        if self.variable_name != [self.name]:
            rep += ', variable_name={}'.format(self.variable_name)
        if self.function:
            rep += ', function={}'.format(self.function)
        
        rep += ')'
        
        return rep


class ScalarOutcome(AbstractOutcome):
    '''
    Scalar Outcome class
    
    Parameters
    ----------
    name : str
           Name of the outcome.
    kind : {INFO, MINIMZE, MAXIMIZE}, optional
    
    Attributes
    ----------
    name : str
    kind : int
    
    '''   
    
    def __init__(self, name, kind=AbstractOutcome.INFO, variable_name=None, 
                 function=None):
        super(ScalarOutcome, self).__init__(name, kind, 
                                            variable_name=variable_name,
                                            function=function)


class TimeSeriesOutcome(AbstractOutcome):
    '''
    TimeSeries Outcome class
    
    Parameters
    ----------
    name : str
           Name of the outcome.
    kind : {INFO, MINIMZE, MAXIMIZE}, optional
    reduce : callable, optional
             a callable which returns a scalar when called. Is only used
             when the outcome is used in an optimization context
    
    Raises
    ------
    ValueError
        if kind is MINIMIZE or MAXIMIZE and callable is not provided or
        not a callable
    
    Attributes
    ----------
    name : str
    kind : int
    reduce : callable
    
    '''   
    
    def __init__(self, name, kind=AbstractOutcome.INFO, variable_name=None, 
                 function=None):
        super(TimeSeriesOutcome, self).__init__(name, kind, variable_name=variable_name, 
                                                function=function)
        
        if (not self.kind==AbstractOutcome.INFO) and (not callable(function)):
            raise ValueError(('function needs to be specified when using'
                              ' TimeSeriesOutcome in optimization' ))

def create_outcomes(outcomes, **kwargs):
    '''Helper function for creating multiple outcomes
    
    Parameters
    ----------
    outcomes : DataFrame, or something convertable to a DataFrame
               in case of string, the string will be passed
    
    Returns
    -------
    list
    
    '''

    if isinstance(outcomes, six.string_types):
        outcomes = pandas.read_csv(outcomes, **kwargs)
    elif not isinstance(outcomes, pandas.DataFrame):
        outcomes = pandas.DataFrame.from_dict(outcomes)
        
    for entry in ['name', 'type']:
        if entry not in outcomes.columns:
            raise ValueError('no {} column in dataframe'.format(entry))
    
    temp_outcomes = []
    for _, row in outcomes.iterrows():
        name = row['name']
        kind = row['type']
        
        if kind=='scalar':
            outcome = ScalarOutcome(name)
        elif kind=='timeseries':
            outcome = TimeSeriesOutcome(name)
        else:
            raise ValueError('unknown type for '+name)
        temp_outcomes.append(outcome)
    return temp_outcomes


'''utilities used throughout em_framework'''
from __future__ import (unicode_literals, print_function, absolute_import,
                        division)

import copy
from collections import OrderedDict

try:
    from UserDict import UserDict
except ImportError:
    from collections import UserDict

try:
    # we assume python 2
    from collections import MutableMapping
except ImportError:
    # so it must by python 3
    from collections.abc import MutableMapping  # @UnusedImport

import itertools
import six

from ..util import EMAError

# from .parameters import Parameter

# Created on Jul 16, 2016
#
# .. codeauthor::jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

__all__ = ['NamedObject', 'NamedDict', 'Counter', 'representation']

class NamedObject(object):

    def __init__(self, name):
        self.name = name


class Counter(object):
    '''helper function for generating counter based names for NamedDicts'''
    
    def __init__(self):
        self._counter = itertools.count()
    
    def __call__(self, named_dict):
        return six.next(self._counter)

    
def representation(named_dict):
    '''helper function for generating repr based names for NamedDicts'''
    return repr(named_dict)


class Variable(NamedObject):
    '''Root class for input parameters and outcomes '''
    
    @property
    def variable_name(self):
        if self._variable_name != None:
            return self._variable_name
        else:
            return [self.name]
        
    @variable_name.setter
    def variable_name(self, name):
        if isinstance(name, six.string_types):
            name = [name]
        self._variable_name = name
        
class NamedObjectMap(object):
     
    def __init__(self, type):  # @ReservedAssignment
        super(NamedObjectMap, self).__init__()
        self.type = type
        self._data = OrderedDict()
         
        if not issubclass(type, NamedObject):
            raise TypeError("type must be a NamedObject")
         
    def clear(self):
        self._data = OrderedDict()
    
    def copy(self):
        copy = NamedObjectMap(self.type)
        copy._data = self._data.copy()
        
        return copy 
        
    def __len__(self):
        return len(self._data)
 
    def __getitem__(self, key):
        if isinstance(key, six.integer_types):
            for i, (_, v) in enumerate(six.iteritems(self._data)):
                if i == key:
                    return v
            raise KeyError(key)
        else:
            return self._data[key]
     
    def __setitem__(self, key, value):
        if not isinstance(value, self.type):
            raise TypeError("can only add " + self.type.__name__ + " objects")
         
        if isinstance(key, six.integer_types):
            self._data = OrderedDict([(value.name, value) if i==key else (k, v) 
                        for i, (k, v) in enumerate(six.iteritems(self._data))])
        else: 
            if value.name != key:
                raise ValueError("key does not match name of " + self.type.__name__)
             
            self._data[key] = value
         
    def __delitem__(self, key):
        del self._data[key]
         
    def __iter__(self):
        return iter(self._data.values())
     
    def __contains__(self, item):
        return item in self._data
     
    def extend(self, value):
        if isinstance(value, NamedObject):
            self._data[value.name] = value
        elif hasattr(value, "__iter__"):
            for item in value:
                self._data[item.name] = item
        else:
            raise TypeError("can only add " + str(type) + " objects")
             
    def __add__(self, value):
        self.extend(value)
        return self
         
    def __iadd__(self, value):
        self.extend(value)
        return self
     
    def keys(self):
        return self._data.keys()

class NamedObjectMapDescriptor(object):
    def __init__(self, kind):
        self.kind = kind
        self.name = None
        self.internal_name = None
        
    def __get__(self, instance, owner):
        if instance is None:
            return self
        try:
            return getattr(instance, self.internal_name)
        except AttributeError:
            map = NamedObjectMap(self.kind)  # @ReservedAssignment
            setattr(instance, self.internal_name, map) 
            return map
 
    def __set__(self, instance, values):
        try:
            map = getattr(instance, self.internal_name)  # @ReservedAssignment
        except AttributeError:
            map = NamedObjectMap(self.kind)  # @ReservedAssignment
            setattr(instance, self.internal_name, map)
        
        map.extend(values)

class NamedDict(UserDict, NamedObject):
    
    def __init__(self, name=representation, **kwargs):
        super(NamedDict, self).__init__(**kwargs)
        if name is None:
            raise ValueError()
        elif callable(name):
            name = name(self)
        self.name = name
           
    
def combine(*args):
    '''combine scenario and policy into a single experiment dict
    
    Parameters
    ----------
    args : two or more dicts that need to be combined
    
    
    Returns
    -------
    a single unified dict containing the entries from all dicts
    
    Raises
    ------
    EMAError 
        if a keyword argument exists in more than one dict
    '''
    experiment = copy.copy(args[0])
    for entry in args[1::]:
        overlap = set(experiment.keys()).intersection(set(entry.keys()))
        if overlap:
            raise EMAError(('parameters exist in two dicts' + str(overlap)))
        experiment.update(entry)

    return experiment

def determine_objects(models, attribute, union=True):
    '''determine the parameters over which to sample
    
    Parameters
    ----------
    models : a collection of AbstractModel instances
    attribute : {'uncertainties', 'levers', 'outcomes'}
    union : bool, optional
            in case of multiple models, sample over the union of
            levers, or over the intersection of the levers
    sampler : Sampler instance, optional
    
    Returns
    -------
    collection of Parameter instances
    
    '''
    try:
        models = iter(models)
    except TypeError:
        # we assume that there is only a single model passed
        models = iter([models])
        
    named_objects = getattr(next(models), attribute).copy()
    intersection = set(named_objects.keys())
    
    # gather parameters across all models
    for model in models:
        model_params = getattr(model, attribute)
        
        # relies on name based identity, do we want that?
        named_objects.extend(model_params)

        intersection = intersection.intersection(model_params.keys())
    
    # in case not union, remove all parameters not in intersection
    if not union:
        params_to_remove = set(named_objects.keys()) - intersection
        for key in params_to_remove:
            del named_objects[key]
    return named_objects



    
    
    


'''

This file illustrated the use of the workbench for doing 
a PRIM analysis.

The data was generated using a system dynamics models implemented in Vensim.
See flu_example.py for the code. The dataset was generated using 32 bit Python.
Therefore, this example will not work if you are running 64 bit Python. 


.. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>
                chamarat <c.hamarat  (at) tudelft (dot) nl>

'''
from __future__ import (division, print_function)
import numpy as np
import matplotlib.pyplot as plt

from ema_workbench import ema_logging, load_results
import ema_workbench.analysis.prim as prim

ema_logging.log_to_stderr(level=ema_logging.INFO)

def classify(data):
    #get the output for deceased population
    result = data['deceased population region 1']
    
    #make an empty array of length equal to number of cases 
    classes =  np.zeros(result.shape[0])
    
    #if deceased population is higher then 1.000.000 people, classify as 1 
    classes[result[:, -1] > 1000000] = 1
    
    return classes

#load data
fn = r'./data/1000 flu cases no policy.tar.gz'
results = load_results(fn)

#perform prim on modified results tuple
prim_obj = prim.setup_prim(results, classify, threshold=0.8, threshold_type=1)

box_1 = prim_obj.find_box()
box_1.show_ppt()
box_1.show_tradeoff()
box_1.inspect(5)
box_1.select(5)
box_1.write_ppt_to_stdout()
box_1.show_pairs_scatter()

#print prim to std_out
print(prim_obj.stats_to_dataframe())
print(prim_obj.boxes_to_dataframe())

#visualize
prim_obj.display_boxes()
plt.show()


'''
Created on 20 dec. 2010

This file illustrated the use of the workbench for a model 
specified in Python itself. The example is based on `Pruyt & Hamarat <http://www.systemdynamics.org/conferences/2010/proceed/papers/P1253.pdf>`_.
For comparison, run both this model and the flu_vensim_no_policy_example.py and 
compare the results. 


.. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>
                chamarat <c.hamarat  (at) tudelft (dot) nl>

'''
from __future__ import (absolute_import, print_function, division, 
                        unicode_literals)

import numpy as np
from numpy import sin, min
from scipy import exp
import matplotlib.pyplot as plt

from ema_workbench import (Model, RealParameter, TimeSeriesOutcome, 
                           perform_experiments, ema_logging)
                         
from ema_workbench.analysis.plotting import lines
from ema_workbench.analysis.plotting_util import KDE                         
from ema_workbench.em_framework.evaluators import MultiprocessingEvaluator


#==============================================================================
#
#    the model itself
#
#==============================================================================

FINAL_TIME = 48
INITIAL_TIME = 0
TIME_STEP = 0.0078125

switch_regions = 1.0
switch_immunity = 1.0
switch_deaths = 1.0
switch_immunity_cap = 1.0

def LookupFunctionX(variable,start,end,step,skew,growth,v=0.5):
    return start + ((end-start)/((1+skew*exp(-growth*(variable-step)))**(1/v)))

def flu_model(x11=0,x12=0,x21=0,x22=0,x31=0,x32=0,x41=0,x51=0,x52=0,x61=0,
                x62=0,x81=0,x82=0,x91=0,x92=0,x101=0,x102=0): 
    #Assigning initial values
    additional_seasonal_immune_population_fraction_R1 = float(x11)
    additional_seasonal_immune_population_fraction_R2 = float(x12)
    
    fatality_rate_region_1 = float(x21)
    fatality_rate_region_2 = float(x22)
    
    initial_immune_fraction_of_the_population_of_region_1 = float(x31)
    initial_immune_fraction_of_the_population_of_region_2 = float(x32)
    
    normal_interregional_contact_rate = float(x41) 
    interregional_contact_rate  = switch_regions*normal_interregional_contact_rate
    
    permanent_immune_population_fraction_R1 = float(x51)
    permanent_immune_population_fraction_R2 = float(x52)
    
    recovery_time_region_1 = float(x61) 
    recovery_time_region_2 = float(x62) 
    
    susceptible_to_immune_population_delay_time_region_1 = 1
    susceptible_to_immune_population_delay_time_region_2 = 1
    
    root_contact_rate_region_1 = float(x81)
    root_contact_rate_region_2 = float(x82)
        
    infection_rate_region_1 = float(x91)
    infection_rate_region_2 = float(x92) 
    
    normal_contact_rate_region_1 = float(x101) 
    normal_contact_rate_region_2 = float(x102) 
    
    ######
    susceptible_to_immune_population_flow_region_1 = 0.0
    susceptible_to_immune_population_flow_region_2 = 0.0
    ######
    
    initial_value_population_region_1 = 6.0*10**8
    initial_value_population_region_2 = 3.0*10**9
    
    initial_value_infected_population_region_1 = 10.0
    initial_value_infected_population_region_2 = 10.0
    
    initial_value_immune_population_region_1 = switch_immunity * initial_immune_fraction_of_the_population_of_region_1 * initial_value_population_region_1
    initial_value_immune_population_region_2 = switch_immunity * initial_immune_fraction_of_the_population_of_region_2 * initial_value_population_region_2
        
    initial_value_susceptible_population_region_1 = initial_value_population_region_1 - initial_value_immune_population_region_1
    initial_value_susceptible_population_region_2 = initial_value_population_region_2 - initial_value_immune_population_region_2
    
    recovered_population_region_1 = 0.0
    recovered_population_region_2 = 0.0
    
    infected_population_region_1 = initial_value_infected_population_region_1
    infected_population_region_2 = initial_value_infected_population_region_2 
    
    susceptible_population_region_1 = initial_value_susceptible_population_region_1 
    susceptible_population_region_2 = initial_value_susceptible_population_region_2
    
    immune_population_region_1 = initial_value_immune_population_region_1
    immune_population_region_2 = initial_value_immune_population_region_2
    
    deceased_population_region_1 = [0.0]
    deceased_population_region_2 = [0.0]
    runTime = [INITIAL_TIME]

    # --End of Initialization--
    
     
    Max_infected = 0

    for time in range(int(INITIAL_TIME/TIME_STEP), int(FINAL_TIME/TIME_STEP)):
        runTime.append(runTime[-1]+TIME_STEP)
        total_population_region_1 = infected_population_region_1 + recovered_population_region_1 + susceptible_population_region_1 + immune_population_region_1
        total_population_region_2 = infected_population_region_2 + recovered_population_region_2 + susceptible_population_region_2 + immune_population_region_2

        infected_population_region_1 = max(0, infected_population_region_1)
        infected_population_region_2 = max(0, infected_population_region_2)

        infected_fraction_region_1 = infected_population_region_1/total_population_region_1
        infected_fraction_region_2 = infected_population_region_2/total_population_region_2
        
        impact_infected_population_on_contact_rate_region_1 = 1-(infected_fraction_region_1**(1/root_contact_rate_region_1))
        impact_infected_population_on_contact_rate_region_2 = 1-(infected_fraction_region_2**(1/root_contact_rate_region_2))
    
#        if ((time*TIME_STEP) >= 4) and ((time*TIME_STEP)<=10):
#            normal_contact_rate_region_1 = float(x101)*(1 - 0.5)
#        else:normal_contact_rate_region_1 = float(x101) 
        
        normal_contact_rate_region_1 = float(x101)*(1 - LookupFunctionX(infected_fraction_region_1, 0, 1, 0.15, 0.75, 15))  
                
        contact_rate_region_1 = normal_contact_rate_region_1*impact_infected_population_on_contact_rate_region_1
        contact_rate_region_2 = normal_contact_rate_region_2*impact_infected_population_on_contact_rate_region_2
        
        recoveries_region_1 =(1-(fatality_rate_region_1*switch_deaths))*infected_population_region_1/recovery_time_region_1
        recoveries_region_2 =(1-(fatality_rate_region_2*switch_deaths))*infected_population_region_2/recovery_time_region_2
        
        flu_deaths_region_1 = fatality_rate_region_1*switch_deaths*infected_population_region_1/recovery_time_region_1
        flu_deaths_region_2 = fatality_rate_region_2*switch_deaths*infected_population_region_2/recovery_time_region_2
        
        infections_region_1 = (susceptible_population_region_1*contact_rate_region_1*infection_rate_region_1*infected_fraction_region_1)+ (susceptible_population_region_1*interregional_contact_rate*infection_rate_region_1*infected_fraction_region_2)
        infections_region_2 = (susceptible_population_region_2*contact_rate_region_2*infection_rate_region_2*infected_fraction_region_2)+ (susceptible_population_region_2*interregional_contact_rate*infection_rate_region_2*infected_fraction_region_1)
            
        infected_population_region_1_NEXT = infected_population_region_1 + (TIME_STEP*(infections_region_1 - flu_deaths_region_1 - recoveries_region_1))
        infected_population_region_2_NEXT = infected_population_region_2 + (TIME_STEP*(infections_region_2 - flu_deaths_region_2 - recoveries_region_2))
    
        if infected_population_region_1_NEXT < 0 or infected_population_region_2_NEXT < 0:
            pass
    
        recovered_population_region_1_NEXT = recovered_population_region_1 + (TIME_STEP*recoveries_region_1)
        recovered_population_region_2_NEXT = recovered_population_region_2 + (TIME_STEP*recoveries_region_2)
        
        if fatality_rate_region_1 >= 0.025:
            qw = 1.0
        elif fatality_rate_region_1 >= 0.01:
            qw = 0.8
        elif fatality_rate_region_1 >= 0.001:
            qw = 0.6
        elif fatality_rate_region_1 >= 0.0001:
            qw = 0.4
        else: qw = 0.2

        if (time*TIME_STEP) <= 10:
            normal_immune_population_fraction_region_1 = (additional_seasonal_immune_population_fraction_R1/2)*sin(4.5+(time*TIME_STEP/2))  + (((2*permanent_immune_population_fraction_R1) + additional_seasonal_immune_population_fraction_R1)/2) 
        else: normal_immune_population_fraction_region_1 = max((float(qw),(additional_seasonal_immune_population_fraction_R1/2)*sin(4.5+(time*TIME_STEP/2))  + (((2*permanent_immune_population_fraction_R1) + additional_seasonal_immune_population_fraction_R1)/2)))

        normal_immune_population_fraction_region_2 = switch_immunity_cap*min((sin((time*TIME_STEP/2)+1.5)*additional_seasonal_immune_population_fraction_R2/2)+(((2*permanent_immune_population_fraction_R2)+additional_seasonal_immune_population_fraction_R2)/2),(permanent_immune_population_fraction_R1+additional_seasonal_immune_population_fraction_R1))+((1-switch_immunity_cap)*((sin((time*TIME_STEP/2)+1.5)*additional_seasonal_immune_population_fraction_R2/2)+(((2*permanent_immune_population_fraction_R2)+additional_seasonal_immune_population_fraction_R2)/2))) 
    
        normal_immune_population_region_1 = normal_immune_population_fraction_region_1*total_population_region_1
        normal_immune_population_region_2 = normal_immune_population_fraction_region_2*total_population_region_2
                
        if switch_immunity == 1:
            susminreg1_1 = ((normal_immune_population_region_1-immune_population_region_1)/susceptible_to_immune_population_delay_time_region_1)
            susminreg1_2 = (susceptible_population_region_1/susceptible_to_immune_population_delay_time_region_1)
            susmaxreg1 = -(immune_population_region_1/susceptible_to_immune_population_delay_time_region_1)
            if (susmaxreg1 >= susminreg1_1) or (susmaxreg1 >= susminreg1_2):
                susceptible_to_immune_population_flow_region_1 = susmaxreg1
            elif (susminreg1_1 < susminreg1_2) and (susminreg1_1 > susmaxreg1):
                susceptible_to_immune_population_flow_region_1 = susminreg1_1
            elif (susminreg1_2 < susminreg1_1) and (susminreg1_2 > susmaxreg1):
                susceptible_to_immune_population_flow_region_1 = susminreg1_2
        else: susceptible_to_immune_population_flow_region_1 = 0
        
        if switch_immunity == 1:
            susminreg2_1 = ((normal_immune_population_region_2-immune_population_region_2)/susceptible_to_immune_population_delay_time_region_2)
            susminreg2_2 = (susceptible_population_region_2/susceptible_to_immune_population_delay_time_region_2)
            susmaxreg2 = -(immune_population_region_2/susceptible_to_immune_population_delay_time_region_2) 
            if (susmaxreg2 >= susminreg2_1) or (susmaxreg2 >= susminreg2_2):
                susceptible_to_immune_population_flow_region_2 = susmaxreg2
            elif (susminreg2_1 < susminreg2_2) and (susminreg2_1 > susmaxreg2):
                susceptible_to_immune_population_flow_region_2 = susminreg2_1
            elif (susminreg2_2 < susminreg2_1) and (susminreg2_2 > susmaxreg2):
                susceptible_to_immune_population_flow_region_2 = susminreg2_2
        else: susceptible_to_immune_population_flow_region_2 = 0
        
        susceptible_population_region_1_NEXT = susceptible_population_region_1 - (TIME_STEP*(infections_region_1 + susceptible_to_immune_population_flow_region_1))
        susceptible_population_region_2_NEXT = susceptible_population_region_2 - (TIME_STEP*(infections_region_2 + susceptible_to_immune_population_flow_region_2))
        
        immune_population_region_1_NEXT = immune_population_region_1 + (TIME_STEP*susceptible_to_immune_population_flow_region_1)
        immune_population_region_2_NEXT = immune_population_region_2 + (TIME_STEP*susceptible_to_immune_population_flow_region_2)
        
        deceased_population_region_1_NEXT = deceased_population_region_1[-1] + (TIME_STEP*flu_deaths_region_1)
        deceased_population_region_2_NEXT = deceased_population_region_2[-1] + (TIME_STEP*flu_deaths_region_2)
        
        #Updating integral values  
        if Max_infected < (infected_population_region_1_NEXT/(infected_population_region_1_NEXT+recovered_population_region_1_NEXT+susceptible_population_region_1_NEXT+immune_population_region_1_NEXT)):
            Max_infected = (infected_population_region_1_NEXT/(infected_population_region_1_NEXT+recovered_population_region_1_NEXT+susceptible_population_region_1_NEXT+immune_population_region_1_NEXT))
        
        recovered_population_region_1 = recovered_population_region_1_NEXT
        recovered_population_region_2 = recovered_population_region_2_NEXT
        
        infected_population_region_1 = infected_population_region_1_NEXT
        infected_population_region_2 = infected_population_region_2_NEXT
        
        susceptible_population_region_1 = susceptible_population_region_1_NEXT
        susceptible_population_region_2 = susceptible_population_region_2_NEXT
    
        immune_population_region_1 = immune_population_region_1_NEXT
        immune_population_region_2 = immune_population_region_2_NEXT
    
        deceased_population_region_1.append(deceased_population_region_1_NEXT)
        deceased_population_region_2.append(deceased_population_region_2_NEXT)
        
        #End of main code
    
    
    return {"TIME":runTime,
            "deceased_population_region_1":deceased_population_region_1} 


if __name__ == '__main__':
    ema_logging.log_to_stderr(ema_logging.INFO)
    
    model = Model('mexicanFlu', function=flu_model)
    model.uncertainties = [RealParameter('x11', 0, 0.5),
                           RealParameter('x12', 0, 0.5),
                           RealParameter('x21', 0.0001, 0.1),
                           RealParameter('x22', 0.0001, 0.1),
                           RealParameter('x31', 0, 0.5),
                           RealParameter('x32', 0, 0.5),
                           RealParameter('x41', 0, 0.9),
                           RealParameter('x51', 0, 0.5),
                           RealParameter('x52', 0, 0.5),
                           RealParameter('x61', 0, 0.8),
                           RealParameter('x62', 0, 0.8),
                           RealParameter('x81', 1, 10),
                           RealParameter('x82', 1, 10),
                           RealParameter('x91', 0, 0.1),
                           RealParameter('x92', 0, 0.1),
                           RealParameter('x101', 0, 200),
                           RealParameter('x102', 0, 200)] 
    
    model.outcomes = [TimeSeriesOutcome("TIME"),
                      TimeSeriesOutcome("deceased_population_region_1")]
    
    nr_experiments = 500
    
    results = perform_experiments(model, 10)
    
#     with MultiprocessingPoolEvaluator(model) as evaluator:
#         results = perform_experiments(model, nr_experiments,
#                                       evaluator=evaluator)
 
    lines(results, outcomes_to_show="deceased_population_region_1", 
          show_envelope=True, density=KDE, titles=None, 
          experiments_to_show=np.arange(0, nr_experiments, 10)
          )
    plt.show()  


'''

Created on Nov 8, 2011

.. codeauthor:: 
     gyucel <g.yucel (at) tudelft (dot) nl>,
     jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>
'''
from __future__ import division

import random

import numpy as np
from ema_workbench.util import info


def distance_same_length(series1, series2, wDim1, wDim2):
    '''
    Calculates the distance between two feature vectors of the same size.
    
    :param series1: Feature vector 1 (2-dimensional numpy array).
    :param series2: Feature vector 2 (2-dimensional numpy array).
    :param wDim1: Weight of the error between the 1st dimensions of the two 
                  feature vectors (i.e. Slope).
    :param wDim2: Weight of the error between the 2nd dimensions of the two 
                  feature vectors (i.e. Curvature).
    
    '''
    
    error = np.square(series1-series2)
    error = np.array([wDim1, wDim2])[np.newaxis].T * error
    error = np.sum(error)
    
    return error/series1.shape[1]

def distance_different_lenght(series1, series2, wDim1, wDim2, sisterCount):
    '''
    Calculates the distance between two feature vectors of different sizes.
    
    :param series1: Feature vector 1 (2-dimensional numpy array).
    :param series2: Feature vector 2 (2-dimensional numpy array).
    :param wDim1: Weight of the error between the 1st dimensions of the two 
                  feature vectors (i.e. Slope).
    :param wDim2: Weight of the error between the 2nd dimensions of the two 
                  feature vectors (i.e. Curvature).
    :param sisterCount: Number of long-versions that will be created for the 
                        short vector.
    '''
    
    # TODO Check why the initial value of error is different for the two 
    # distance methods?
    
    length1 = series1.shape[1]
    length2 = series2.shape[1]
#    toAdd = abs(length1-length2)
    
    if length1>length2:
        shortFV = series2
        longFV = series1
    else:
        shortFV = series1
        longFV = series2
    
    sisters = create_sisters(shortFV, longFV.shape, sisterCount)
    
    # to take advantage of the fact that the sisters are in a 3d array
    # I also vectorized the error calculation.
    # this means that calculation time is almost independent from the number
    # of sisters you want to use.
    error = np.square(sisters - longFV.T[np.newaxis,:,:])
    weights = np.array([wDim1, wDim2])
    
    error = error*weights[np.newaxis, np.newaxis, :]
    
    # this is sort of stupid, it should be possible to do in one line
    # but axis=(1,2) does not work, it requires numpy 1.7. I have 1.6.1 
    # at the moment
    error = np.sum(error, axis=1)
    error = np.sum(error, axis=1)
    return np.min(error)/longFV.shape[1]
#    for i in range(sisterCount):
#        sisterFV = createSister(shortFV, toAdd)
#        tempError = distance_same_length(sisterFV, longFV, wDim1, wDim2)
#        if tempError < error:
#            error = tempError
#    return error

def create_sisters(shortFV, desired_shape, sisterCount):
    '''
    Creates a new feature vector that is behaviorally identical to the given 
    vector by adding toAdd number of segments.
    
    :param shortFV: The feature vector to be extended.
    :param toAdd: Number of sections to be added to the input vector while 
                  creating the equivalent sister.
    ''' 
    
    #determine how much longer the vector has to become
    to_add = desired_shape[1]-shortFV.shape[1]
    
    #create a 2d array of indices
    indices = np.zeros(shape=(sisterCount, desired_shape[1]),dtype=int)
    
    #fill the first part of the indices array with random numbers
    #these are the indices that will be used to extent the short vector
    indices[:, 0:to_add] = np.random.randint(0, 
                                             shortFV.shape[1], 
                                             size=(sisterCount, to_add))
    
    #add the indices for the full vector to the rest
    indices[:, to_add::] = np.arange(0, shortFV.shape[1])[np.newaxis, :]
    
    #sort indices
    indices = np.sort(indices, axis=1)
    
    #this is where the real magic happens, we use the generated indices
    #in order to generate in one line of code all the sisters
    sisters = shortFV.T[indices,:] 
    
    return sisters


def createSister(shortFV, toAdd):
    '''
    Creates a new feature vector that is behaviorally identical to the given 
    vector by adding toAdd number of segments.
    
    :param shortFV: The feature vector to be extended.
    :param toAdd: Number of sections to be added to the input vector while 
                  creating the equivalent sister.
    ''' 
    sister = np.zeros(shape=(shortFV.shape[0], shortFV.shape[1]+toAdd))
    
    
    # while you have to add, add a random number of entries
    index = 0 
    i = 0
    while (toAdd>0) and (i<shortFV.shape[1]):
        
        x = random.randint(0, toAdd)
        sister[:, index:index+x+1] = shortFV[:,i][np.newaxis].T
        toAdd -= x
        index += x+1
        i+=1
    
    #fill up with the remaining values from short
    if i<shortFV.shape[1]:
        sister[:, index::] = shortFV[:, i::]

    return sister

def construct_features(data, filterSlope, tHoldSlope, filterCurvature, 
                       tHoldCurvature, addMidExtension, addEndExtension):
    '''
    Constructs a feature vector for each of the data-series contained in the 
    data. 
    
    '''
    info("calculating features")
    
    # TODO, the casting of each feature to a list of tuples might be 
    # removed at some stage, it will lead to a speed up, for you 
    # can vectorize the calculations that use the feature vector
    features = []
    for i in range(data.shape[0]):
        feature = construct_feature_vector(data[i, :], filterSlope, tHoldSlope, 
                                     filterCurvature, tHoldCurvature, 
                                     addMidExtension, addEndExtension)
#        feature =  [tuple(feature[0,:]),tuple(feature[1,:])]
        features.append(feature)
    return features

def filter_series(series, parentSeries, thold):
    '''
    Filters out a given time-series for insignificant fluctuations. For 
    example very small fluctuations due to numeric error of the simulator).
    '''
    absParent = np.absolute(parentSeries[0:parentSeries.shape[0]-1])
    absSeries = np.absolute(series)
    cond1a = np.not_equal(absParent, 0)
    cond1b = absSeries < thold*absParent
    cond2a = np.logical_not(cond1b)
    cond2b = absSeries < thold/10
    cond1 = np.logical_and(cond1a,cond1b)
    cond2 = np.logical_and(cond2a,cond2b)
    cond = np.logical_or(cond1,cond2)
    series[cond] = 0
    return series

def extend_mids(vector):
    sections = vector[0].size
    added = 0
    for i in range(sections-1):
        if(vector[0][i+added]*vector[0][i+1+added]==-1) and\
          (vector[0][i+added+1]!=0):
            vector = np.insert(vector, i+1+added, 0, axis=1)
            vector[1][i+1+added] = vector[1][i+added]
            added+=1
    return vector

def extend_ends(vector):
    sections = vector[0].size
    added = 0
    
    if(vector[0][0]*vector[1][0]==1):
        #Front extension
        vector = np.insert(vector, 0, 0, axis=1)
        vector[1][0] = vector[1][1]
        added+=1
    if(vector[0][sections-1+added]*vector[1][sections-1+added]==-1):
        #End extension
        vector = np.append(vector, [[0],[0]], axis=1)
        vector[1][sections+added] = vector[1][sections-1+added]
        added+=1
    return vector

def construct_feature_vector(dataSeries, filterSlope, tHoldSlope, 
                             filterCurvature, tHoldCurvature, addMidExtension, 
                             addEndExtension):
    '''
    Constructs a feature vector for the given dataSeries. Each element in 
    this 2-D vector represents a section along the time-series that can be 
    characterized as an atomic behaviour mode.
    '''
    dsLength = dataSeries.shape[0]
    slope = dataSeries[1::]-dataSeries[0:dsLength-1]
    curvature = slope[1::]-slope[0:dsLength-2]
    
    if filterSlope:
        slope = filter_series(slope, dataSeries, tHoldSlope)
    if filterCurvature:
        curvature = filter_series(curvature, slope, tHoldCurvature)
                
    signSlope = slope.copy()
    signSlope[signSlope>0] = 1
    signSlope[signSlope<0] = -1
    signSlope[signSlope==0] = 0
    signSlope = np.delete(signSlope, -1)
    
    signCurvature = curvature.copy()
    signCurvature[signCurvature>0] = 1
    signCurvature[signCurvature<0] = -1
    signCurvature[signCurvature==0] = 0         
    
    sections = 10*signSlope+signCurvature
    temp = sections[1::]-sections[0:sections.shape[0]-1]
    transPoints = np.nonzero(temp)
    numberOfSections = len(transPoints[0])+1
    
    featureVector = np.zeros(shape=(2,numberOfSections))
    
    for k in transPoints:
        featureVector[0][0:len(featureVector[0])-1] = signSlope[k]
        featureVector[1][0:len(featureVector[0])-1] = signCurvature[k]
    featureVector[0][numberOfSections-1]= signSlope[-1]
    featureVector[1][numberOfSections-1]= signCurvature[-1]
 
    
    if addMidExtension:
        featureVector = extend_mids(featureVector)
    if addEndExtension:
        featureVector = extend_ends(featureVector)
    return featureVector

def distance_gonenc(data,
                    sisterCount=50, 
                    wSlopeError=1, 
                    wCurvatureError=1,
                    filterSlope=True,
                    tHoldSlope = 0.1,
                    filterCurvature=True,
                    tHoldCurvature=0.1,
                    addMidExtension=True,
                    addEndExtension=True
                    ):
    
    '''
    The distance measures the proximity of data series in terms of their 
    qualitative pattern features. In order words, it quantifies the proximity 
    between two different dynamic behaviour modes.
    
    It is designed to work mainly on non-stationary data. It's current version 
    does not perform well in catching the proximity of two cyclic/repetitive 
    patterns with different number of cycles (e.g. oscillation with 4 cycle 
    versus oscillation with 6 cycles).
    
    :param data:
    :param sisterCount: Number of long-versions that will be created for the 
                        short vector while comparing two data series with 
                        unequal feature vector lengths. 
    :param wSlopeError: Weight of the error between the 1st dimensions of the 
                        two feature vectors (i.e. Slope). (default=1)
    :param wCurvatureError: Weight of the error between the 2nd dimensions of 
                            the two feature vectors (i.e. Curvature). 
                            (default=1)
    :param wFilterSlope: Boolean, indicating whether the slope vectors should 
                         be filtered for minor fluctuations, or not. 
                         (default=True)
    :param tHoldSlope: The threshold value to be used in filtering out 
                       fluctuations in the slope. (default=0.1)
    :param filterCurvature: Boolean, indicating whether the curvature vectors 
                            should be filtered for minor fluctuations, or not.
                            (default=True)
    :param tHoldCurvature: The threshold value to be used in filtering out 
                           fluctuations in the curvature. (default=0.1)
    :param addMidExtension: Boolean, indicating whether the feature vectors 
                            should be extended by introducing transition 
                            sections along the vector.
                            (default=True)
    :param addEndExtension: Boolean, indicating whether the feature vectors 
                            should be extended by introducing startup/closing 
                            sections at the beginning/end of the vector.
                            (default=True)
    '''
    
    
    runLogs = []
    #Generates the feature vectors for all the time series that are contained 
    # in numpy array data
    features = construct_features(data, filterSlope, tHoldSlope, 
                                  filterCurvature, tHoldCurvature, 
                                  addMidExtension, addEndExtension)
    info("calculating distances")
    dRow = np.zeros(shape=(np.sum(np.arange(data.shape[0])), ))
    index = -1
    for i in range(data.shape[0]):
        feature_i = features[i]
            
        # For each run, a log is created
        # Log includes a description dictionary that has key information 
        # for post-clustering analysis, and the data series itself. These 
        # logs are stored in a global array named runLogs
        behaviorDesc = {}
        behaviorDesc['Index'] = str(i)
        
        #this may not work due to data type mismatch
        featVector = feature_i
        
        behaviorDesc['Feature vector'] = str(featVector)
        behavior = data[i]
        localLog = (behaviorDesc, behavior)
        runLogs.append(localLog)
    
        for j in range(i+1, data.shape[0]):
            index += 1
            feature_j = features[j]
            if feature_i.shape[1] == feature_j.shape[1]:
                distance = distance_same_length(feature_i, feature_j, 
                                                wSlopeError, wCurvatureError)
    
            else:
                distance = distance_different_lenght(feature_i, 
                                                     feature_j, 
                                                     wSlopeError, 
                                                     wCurvatureError, 
                                                     sisterCount)
            dRow[index] = distance
    return dRow, runLogs


if __name__ == '__main__':
    tester = np.array([(0, 1, 4, 8,16,24,30,34,36,39,34,38,34)])
    #tester = np.array([(0, 3, 10,8,16,24,30,34,36,39,34,30,29),(0, 3, 10,8,16,24,30,34,36,39,34,30,29),(0, 3, 10,8,16,24,30,34,36,39,34,30,29)])
    distance_gonenc(tester, addMidExtension=True, addEndExtension=True)
    
    


'''
A scenario discovery oriented implementation of CART. It essentially is a 
wrapper around scikit-learn's version of CART. 


'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

import math

import numpy as np
import numpy.lib.recfunctions as recfunctions
import six
from sklearn import tree
from sklearn.externals.six import StringIO

from . import scenario_discovery_util as sdutil
from ..util import ema_logging

# Created on May 22, 2015
# 
# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>


__all__ = ['setup_cart',
           'CART']

def setup_cart(results, classify, incl_unc=[], mass_min=0.05):
    """helper function for performing cart in combination with data
    generated by the workbench. 
    
    Parameters
    ----------
    results : tuple of structured array and dict with numpy arrays
              the return from :meth:`perform_experiments`.
    classify : string, function or callable
               either a string denoting the outcome of interest to 
               use or a function. 
    incl_unc : list of strings
    mass_min : float
    
    
    Raises
    ------
    TypeError 
        if classify is not a string or a callable.
    
    """
    
    if not incl_unc:
        x = np.ma.array(results[0])
    else:
        drop_names = set(recfunctions.get_names(results[0].dtype))-set(incl_unc)
        x = recfunctions.drop_fields(results[0], drop_names, asrecarray = True)
    if isinstance(classify, six.string_types):
        y = results[1][classify]
        mode = sdutil.REGRESSION
    elif callable(classify):
        y = classify(results[1])
        mode = sdutil.BINARY
    else:
        raise TypeError("unknown type for classify")
    
    return CART(x, y, mass_min, mode=mode)


class CART(sdutil.OutputFormatterMixin):
    '''CART algorithm
    
    can be used in a manner similar to PRIM. It provides access
    to the underlying tree, but it can also show the boxes described by the
    tree in a table or graph form similar to prim.

    Parameters
    ----------
    x : recarray
    y : ndarray
    mass_min : float, optional
               a value between 0 and 1 indicating the minimum fraction
               of data points in a terminal leaf. Defaults to 0.05, 
               identical to prim. 
    mode : {BINARY, CLASSIFICATION, REGRESSION}
           indicates the mode in which CART is used. Binary indicates
           binary classification, classification is multiclass, and regression
           is regression.

    Attributes
    ----------
    boxes : list
            list of recarray box lims
    stats : list
            list of dicts with stats 
    
    Notes
    -----
    This class is a wrapper around scikit-learn's CART algorithm. It provides
    an interface to CART that is more oriented towards scenario discovery, and
    shared some methods with PRIM
    
    See also
    --------
    :mod:`prim`
        
    '''
    
    sep = '?!?'
    
    def __init__(self, x,y, mass_min=0.05, mode=sdutil.BINARY):
        ''' init
        
        '''
        x = recfunctions.drop_fields(x, "scenario_id", asrecarray=True)

        
        self.x = x
        self.y = y
        self.mass_min = mass_min
        self.mode = mode

        # we need to transform the structured array to a ndarray
        # we use dummy variables for each category in case of categorical 
        # variables. Integers are treated as floats
        self.feature_names = []
        columns = []
        for unc, dtype in x.dtype.descr:
            dtype = x.dtype.fields[unc][0]
            if dtype==np.object:
                categories =  sorted(list(set(x[unc])))
                for cat in categories:
                    label = '{}{}{}'.format(unc, self.sep,cat)
                    self.feature_names.append(label)
                    columns.append(x[unc]==cat)
            else:
                self.feature_names.append(unc)
                columns.append(x[unc])

        self._x = np.column_stack(columns)
        self._boxes = None
        self._stats = None

    @property
    def boxes(self):
        assert self.clf
        
        if self._boxes:
            return self._boxes
    
        # based on
        # http://stackoverflow.com/questions/20224526/how-to-extract-the-
        # decision-rules-from-scikit-learn-decision-tree
        assert self.clf
        
        left = self.clf.tree_.children_left
        right = self.clf.tree_.children_right
        threshold = self.clf.tree_.threshold
        features = [self.feature_names[i] for i in self.clf.tree_.feature]
    
        # get ids of leaf nodes
        leafs = np.argwhere(left == -1)[:,0]     
    
        def recurse(left, right, child, lineage=None):          
            if lineage is None:
                # lineage = [self.clf.tree_.value[child]]
                lineage = []
            
            if child in left:
                parent = np.where(left == child)[0].item()
                split = 'l'
            else:
                parent = np.where(right == child)[0].item()
                split = 'r'
    
            lineage.append((parent, split, threshold[parent],
                            features[parent]))
    
            if parent == 0:
                lineage.reverse()
                return lineage
            else:
                return recurse(left, right, parent, lineage)
            
        box_init = sdutil._make_box(self.x)
        boxes = []
        for leaf in leafs:
            branch = recurse(left, right, leaf)
            box = np.copy(box_init)
            for node in branch:
                direction = node[1]
                value = node[2]
                unc = node[3]
                
                if direction=='l':
                    try:
                        box[unc][1] = value
                    except ValueError:
                        unc, cat = unc.split(self.sep)
                        cats = list(box[unc][0])
                        cats = [str(cat) for cat in cats]
                        cats.pop(cats.index(str(cat)))
                        box[unc][:]=set(cats)
                else:
                    try:
                        if (box.dtype.fields[unc][0])==np.int32:
                            value = math.ceil(value)
                        box[unc][0] = value
                    except (ValueError, KeyError):
                        # we are in the right hand branch, so 
                        # the category is included
                        pass
                        
            boxes.append(box) 
        self._boxes = boxes
        return self._boxes       
    
    @property
    def stats(self):
        if self._stats:
            return self._stats
        
        boxes = self.boxes
        
        box_init = sdutil._make_box(self.x)
        
        self._stats = []
        for box in boxes:
            boxstats = self._boxstat_methods[self.mode](self, box, box_init)
            self._stats.append(boxstats)
        return self._stats

    
    def _binary_stats(self, box, box_init):
        indices = sdutil._in_box(self.x, box)
            
        y_in_box = self.y[indices]
        box_coi = np.sum(y_in_box)
        
        boxstats = {'coverage': box_coi/np.sum(self.y),
                    'density': box_coi/y_in_box.shape[0],
                    'res dim':sdutil._determine_nr_restricted_dims(box,
                                                                   box_init),
                    'mass':y_in_box.shape[0]/self.y.shape[0]}
        return boxstats
    
    def _regression_stats(self, box, box_init):
        indices = sdutil._in_box(self.x, box)
            
        y_in_box = self.y[indices]
        
        boxstats = {'mean': np.mean(y_in_box),
                    'mass':y_in_box.shape[0]/self.y.shape[0],
                    'res dim':sdutil._determine_nr_restricted_dims(box,
                                                                   box_init)}
        return boxstats

    
    def _classification_stats(self, box, box_init):
        indices = sdutil._in_box(self.x, box)
            
        y_in_box = self.y[indices]
        classes = set(self.y)
        classes = list(classes)
        classes.sort()
        
        counts = [y_in_box[y_in_box==ci].shape[0] for ci in classes]

        total_gini = 0
        for count in counts:
            total_gini += (count/y_in_box.shape[0])**2
        gini = 1 - total_gini
        
        boxstats = {'gini': gini,
            'mass':y_in_box.shape[0]/self.y.shape[0],
            'box_composition': counts,
            'res dim':sdutil._determine_nr_restricted_dims(box,
                                                           box_init)}
        
        return boxstats

    _boxstat_methods = {sdutil.BINARY:_binary_stats, 
                        sdutil.REGRESSION:_regression_stats,
                        sdutil.CLASSIFICATION: _classification_stats}

    def build_tree(self):
        '''train CART on the data'''
        min_samples = int(self.mass_min*self.x.shape[0])
        
        if self.mode==sdutil.REGRESSION:
            self.clf =  tree.DecisionTreeRegressor(min_samples_leaf=min_samples)
        else:
            self.clf = tree.DecisionTreeClassifier(min_samples_leaf=min_samples)
        self.clf.fit(self._x,self.y)

    def show_tree(self):
        '''return a png of the tree'''
        assert self.clf
        try:
            import pydotplus as pydot
        except ImportError:
            import pydot # dirty hack for read the docs

        dot_data = StringIO() 
        tree.export_graphviz(self.clf, out_file=dot_data, 
                             feature_names=self.feature_names) 
        graph = pydot.graph_from_dot_data(dot_data.getvalue().encode('ascii'))  # @UndefinedVariable
        img = graph.create_png()
        return img

       
# if __name__ == '__main__':
#     from test import test_utilities
#     import matplotlib.pyplot as plt
# 
#     ema_logging.log_to_stderr(ema_logging.INFO)
# 
#     def scarcity_classify(outcomes):
#         outcome = outcomes['relative market price']
#         change = np.abs(outcome[:, 1::]-outcome[:, 0:-1])
#         
#         neg_change = np.min(change, axis=1)
#         pos_change = np.max(change, axis=1)
#         
#         logical = (neg_change > -0.6) & (pos_change > 0.6)
#         
#         classes = np.zeros(outcome.shape[0])
#         classes[logical] = 1
#         
#         return classes
#  
#     results = test_utilities.load_scarcity_data()
#     
#     cart = setup_cart(results, scarcity_classify)
#     cart.build_tree()
#     
#     print(cart.boxes_to_dataframe())
#     print(cart.stats_to_dataframe())
#     cart.display_boxes(together=True)
#     
#     img = cart.show_tree()
#      
#     import matplotlib.pyplot as plt
#     import matplotlib.image as mpimg
#   
#     # treat the dot output string as an image file
#     sio = StringIO()
#     sio.write(img)
#     sio.seek(0)
#     img = mpimg.imread(sio)
#       
#     # plot the image
#     imgplot = plt.imshow(img, aspect='equal')
#       
#     plt.show()


'''
Created on Sep 27, 2011

@author: gyucel
'''
from __future__ import (print_function, unicode_literals)
import numpy as np
from ema_workbench.util import utilities

def mean(ds):
    return np.mean(ds)

def variance(ds):
    return np.var(ds)

def stdDev(ds):
    return np.std(ds)

def linearFit(ds):
    return nthDegreeFit(range(len(ds)),ds,1)

def quadraticFit(ds):
    return nthDegreeFit(range(len(ds)),ds,2)

def nthDegreeFit(x,ds,degree):
    return np.polyfit(x, ds, degree)


def exponentialFit(ds):
    return 0

def autoCovariance(ds,lag):
    N = len(ds)
    k = lag
    m = mean(ds)
    sum = 0
    for i in range(N-k):
        sum += (ds[i]-m)*(ds[i+k]-m)
    return sum/N

def autoCorrelation(ds, lag):
    k = np.abs(lag)
    return autoCovariance(ds,k)/autoCovariance(ds,0)

def crossCorrelation(ds1, ds2, lag):
    std1 = stdDev(ds1)
    std2 = stdDev(ds2)
    m1 = mean(ds1)
    m2 = mean(ds2)
    K = 0
    for i in range(len(ds1)-lag):
        K+= (ds1[i+lag]-m1)*(ds2[i]-m2)
    crossCorr = (K/len(ds1))/(std1*std2)
    return crossCorr   

def varAutoCorrelation(ds, lag):
    N = len(ds)
    k = lag
    sum = 0
    for j in range(N-2):
        i = j+1
        a = autoCorrelation(ds,k-i)+autoCorrelation(ds,k+i)-2*autoCorrelation(ds,k)*autoCorrelation(ds,i)
        aSqr = a*a
        sum += (N-i)*aSqr
    var = sum/(N*(N+2))
    return var

def periodDominance(ds):
    Y = np.fft.rfft(ds)
    n = len(Y)
    powerSpect = np.abs(Y)**2
    timeStep = 1 
    freq = np.fft.fftfreq(n, d=timeStep)
    print(len(freq), len(powerSpect))
    for i in range(len(freq)/2+1):
        print(freq[i], 1/freq[i], powerSpect[i])


if __name__ == '__main__':
    
    cases, results = utilities.load_results('PatternSet_Periodic.cpickle')
    dataSeries = results.get('outcome')
    ds1 = dataSeries[25]
    ds2 = dataSeries[26]
    
    print(linearFit(ds1))
    print(quadraticFit(ds1))
    print(mean(ds1), variance(ds1), stdDev(ds1))
    print(autoCovariance(ds1,0))
    for k in range(31):
        print(k,autoCorrelation(ds1,k))
    
    for k in range(31):
        print(k, crossCorrelation(ds1,ds2,k))
    
    periodDominance(ds1)    
#    data = np.zeros(5,dtype=int)
#    data[0] = 2
#    data[1] = 4
#    data[2] = 6
#    data[3] = 8
#    data[4] = 10
#    print mean(data), varianqce(data), stdDev(data)
#    print autoCovariance(data,2)
#    print autoCorrelation(data,0),autoCorrelation(data,2)
#    linearFit(data)
#    r = nthDegreeFit(data,0)
#    print r


'''
This module specifies the abstract base class for interfacing with models. 
Any model that is to be controlled from the workbench is controlled via
an instance of an extension of this abstract base class. 

'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

import abc
import copy
import operator
import os
import six
import warnings
from ema_workbench.util import ema_logging
from ema_workbench.em_framework.parameters import CategoricalParameter


try:
    from collections import MutableMapping
except ImportError:
    from collections.abc import MutableMapping  # @UnusedImport

from collections import defaultdict

from .util import (NamedObject, combine, NamedObjectMapDescriptor)
from .parameters import Parameter, Constant
from .outcomes import AbstractOutcome

from ..util import debug, EMAError
from ..util.ema_logging import method_logger

# Created on 23 dec. 2010
# 
# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>
# 

__all__ = ['AbstractModel', 'Model', 'FileModel', 'Replicator', 
           'SingleReplication']

#==============================================================================
# abstract Model class 
#==============================================================================
class ModelMeta(abc.ABCMeta):
    
    def __new__(mcls, name, bases, namespace):  # @NoSelf
        
        for key, value in namespace.items():
            if isinstance(value, NamedObjectMapDescriptor):
                value.name = key
                value.internal_name = '_'+key
       
        return abc.ABCMeta.__new__(mcls, name, bases, namespace)


class AbstractModel(six.with_metaclass(ModelMeta, NamedObject)):
    '''
    :class:`ModelStructureInterface` is one of the the two main classes used 
    for performing EMA. This is an abstract base class and cannot be used 
    directly. When extending this class :meth:`model_init` and 
    :meth:`run_model` have to be implemented. 
    
    
    Attributes
    ----------
    uncertainties : listlike
                    list of parameter 
    levers : listlike
             list of parameter instances
    outcomes : listlike
               list of outcome instances
    name : str
           alphanumerical name of model structure interface
    output : dict
             this should be a dict with the names of the outcomes as key
    
    '''
    
    @property
    def output(self):
        return self._output

    @output.setter
    def output(self, outputs):
        for outcome in self.outcomes:
            data = [outputs[var] for var in outcome.variable_name]
            self._output[outcome.name] = outcome.process(data)
    
    @property
    def outcome_variables(self):
        if self._outcome_variables is None:
            self._outcome_variables = [var for o in self.outcomes for var in 
                                       o.variable_name]
        return self._outcome_variables
         
    uncertainties = NamedObjectMapDescriptor(Parameter)
    levers = NamedObjectMapDescriptor(Parameter)
    outcomes = NamedObjectMapDescriptor(AbstractOutcome)
    constants = NamedObjectMapDescriptor(Constant)
        
    def __init__(self, name):
        """
        interface to the model
        
        Parameters
        ----------
        name : str
               name of the modelInterface. The name should contain only
               alpha-numerical characters.        

               
        Raises
        ------
        EMAError if name contains non alpha-numerical characters
        
        """
        super(AbstractModel, self).__init__(name)

        if not self.name.isalnum():
            raise EMAError("name of model should only contain alpha numerical\
                            characters")

        self._outcome_variables = None
        self._output = {}
        
    @method_logger
    def model_init(self, policy):
        '''
        Method called to initialize the model.
        
        Parameters
        ----------
        policy : dict
                 policy to be run.
 
        
        Note
        ----
        This method should always be implemented. Although in simple cases, a 
        simple pass can suffice.
         
        '''
        self.policy = policy

        remove = []
        for key, value in policy.items():
            if hasattr(self, key):
                setattr(self, key, value)
                remove.append(key)
        
        for k in remove:
            del policy[k]
                
        

    def _transform(self, sampled_parameters, parameters):
        
        if not parameters:
            # no parameters defined, so nothing to transform, mainly
            # useful for manual specification of scenario /  policy
            # without having to define uncertainties / levers
            return
        
        temp = {}
        for par in parameters:
            # only keep uncertainties that exist in this model
            try:
                value = sampled_parameters[par.name]
            except KeyError:
                if par.default is not None:
                    value = par.default
                else:
                    ema_logging.debug('{} not found'.format(par.name))
                    continue

            multivalue = False
            if isinstance(par, CategoricalParameter):
                if par.multivalue == True:
                    multivalue = True
                    values = value
            
            # translate uncertainty name to variable name
            for i, varname in enumerate(par.variable_name):
                # a bit hacky implementation, investigate some kind of 
                # zipping of variable_names and values
                if multivalue:
                    value = values[i]
                
                temp[varname] = value
        
        sampled_parameters.data = temp

    @method_logger    
    def run_model(self, scenario, policy):
        """
        Method for running an instantiated model structure. 
        
        Parameters
        ----------
        scenario : Scenario instance
        policy : Policy instance
        
        
        """
        if not self.initialized(policy):
            self.model_init(policy)
        
        #TODO:: here we need to add policies and constants in some manner
        self._transform(scenario, self.uncertainties)
        self._transform(policy, self.levers)

    @method_logger
    def initialized(self, policy):
        '''check if model has been initialized 

        Parameters
        ----------
        policy : a Policy instance
        
        '''
        
        try:
            return self.policy.name == policy.name
        except AttributeError:
            return False

    @method_logger
    def retrieve_output(self):
        """
        Method for retrieving output after a model run.
        
        Returns
        -------
        dict with the results of a model run. 
        """
        warnings.warn('deprecated, use model.output instead')
        return self.output
    
    @method_logger
    def reset_model(self):
        """
        Method for reseting the model to its initial state. The default
        implementation only sets the outputs to an empty dict. 

        """
        self._output = {}
    
    @method_logger
    def cleanup(self):
        '''
        This model is called after finishing all the experiments, but 
        just prior to returning the results. This method gives a hook for
        doing any cleanup, such as closing applications. 
        
        In case of running in parallel, this method is called during 
        the cleanup of the pool, just prior to removing the temporary 
        directories. 
        
        '''
        pass
    
    def as_dict(self):
        '''returns a dict representation of the model'''
        
        def join_attr(field):
            joined = ', '.join([repr(entry) for entry in 
                                sorted(field, key=operator.attrgetter('name'))])
            return '[{}]'.format(joined)
        model_spec = {}
        
        klass = self.__class__.__name__
        name = self.name
                
        uncs = ''
        for uncertainty in self.uncertainties:
            uncs += '\n' + repr(uncertainty)
            
        model_spec['class'] = klass
        model_spec['name'] = name
        model_spec['uncertainties'] = join_attr(self.uncertainties)
        model_spec['outcomes'] = join_attr(self.outcomes)
        model_spec['constants'] = join_attr(self.constants)
            
        return model_spec

class Replicator(AbstractModel):
    @property
    def replications(self):
        return self._replications
    
    @replications.setter
    def replications(self, replications):
        
        # int
        if isinstance(replications, int):
            # TODO:: use a repeating generator instead
        
            self._replications = [{} for _ in range(replications)]
            self.nreplications = replications
        elif isinstance(replications, list):
            # should we check if all are dict?
            self._replications = replications
            self.nreplications = len(replications) 
        else:
            raise TypeError("replications should be int or list not {}".format(type(replications)))

    @method_logger
    def run_model(self, scenario, policy):
        """
        Method for running an instantiated model structure. 
        
        Parameters
        ----------
        scenario : Scenario instance
        policy : Policy instance
        
        
        """
        super(Replicator, self).run_model(scenario, policy)
        
        constants = {c.name:c.value for c in self.constants}
        
        outputs = defaultdict(list)
        
        policy = copy.deepcopy(self.policy)
        partial_experiment = combine(scenario, policy, constants)
        
        for _, rep in enumerate(self.replications):
            experiment = copy.deepcopy(partial_experiment)
            experiment = combine(experiment, rep)
            output = self.run_experiment(experiment)
            for key, value in output.items():
                outputs[key].append(value)
            
        self.output = outputs


class SingleReplication(AbstractModel):
    @method_logger
    def run_model(self, scenario, policy):
        """
        Method for running an instantiated model structure. 
        
        Parameters
        ----------
        scenario : Scenario instance
        policy : Policy instance
        
        
        """
        super(SingleReplication, self).run_model(scenario, policy) 
        # TODO:: should this not be moved up?
        constants = {c.name:c.value for c in self.constants}
        experiment = combine(scenario, self.policy, constants)
        self.output = self.run_experiment(experiment)


class BaseModel(AbstractModel):
    '''
    generic class for working with models implemented as a Python callable 
    
    Parameters
    ----------
    name : str
    function : callable
               a function with each of the uncertain parameters as a keyword
               argument
    
    
    Attributes
    ----------
    uncertainties : listlike
                    list of parameter 
    levers : listlike
             list of parameter instances
    outcomes : listlike
               list of outcome instances
    name : str
           alphanumerical name of model structure interface
    output : dict
             this should be a dict with the names of the outcomes as key
    working_directory : str
                        absolute path, all file operations in the model
                        structure interface should be resolved from this
                        directory. 
    
    '''

    
    def __init__(self, name, function=None):
        super(BaseModel, self).__init__(name)
        
        if not callable(function):
            raise ValueError('function should be callable')
        
        self.function = function

    @method_logger
    def run_experiment(self, experiment):
        """
        Method for running an instantiated model structure. 
        
        Parameters
        ----------
        experiment : dict like
        
        
        """
        model_output = self.function(**experiment)
        
        # TODO: might it be possible to somehow abstract this
        # perhaps expose a get_data on modelInterface?
        # different connectors can than implement only this
        # get method
        results  = {}
        for i, variable in enumerate(self.outcome_variables):
            try:
                value = model_output[variable]
            except KeyError:
                ema_logging.warning(variable +' not found in model output')
                value  = None
            except TypeError:
                value = model_output[i]
            results[variable] = value
        return results
        
    def as_dict(self):
        model_specs = super(BaseModel, self).as_dict()
        model_specs['function'] = self.function
        return model_specs


# TODO:: is it required that FileModel has a model_file?
# might instead be some data files used by a Python Process
# Solution, have an intermediate class, WorkingDirectoryModel
# FileModel than extents that class with the model_file
class FileModel(AbstractModel):
    @property
    def working_directory(self):
        return self._working_directory
    
    @working_directory.setter
    def working_directory(self, path):
        wd = os.path.abspath(path)
        debug('setting working directory to '+ wd)
        self._working_directory = wd

    def __init__(self, name, wd=None, model_file=None):
        """interface to the model
        
        interface to the model
        
        Parameters
        ----------
        name : str
               name of the modelInterface. The name should contain only
               alpha-numerical characters.        
        working_directory : str
                            working_directory for the model. 
        model_file  : str
                     The model file relative to working directory
               
        Raises
        ------
        EMAError 
            if name contains non alpha-numerical characters
        ValueError
            if model_file cannot be found
        
        """
        super(FileModel, self).__init__(name)
        self.working_directory = wd

        path_to_file = os.path.join(self.working_directory, model_file)
        if not os.path.isfile(path_to_file):
            raise ValueError('cannot find model file')
        
        self.model_file = model_file

    def as_dict(self):
        model_specs = super(FileModel, self).as_dict()
        model_specs['model_file'] = self.model_file
        model_specs['working_directory'] = self.working_directory
        return model_specs

class Model(SingleReplication, BaseModel):
    pass